#  ==================================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ==================================================================================

import logging
logger = logging.getLogger('camelot.view.export.excel')

from pyExcelerator import *
import datetime

titleFont = Font()              # initializing titleFont Object
headerFont = Font()             # initializing headerFont Object
cellFont = Font()               # initializing cellFont Object

titleFont.name = 'Arial'        # Setting Fonts Name
headerFont.name = 'Arial'
cellFont.name = 'Arial'

titleFont.bold = True           # Setting title font to bold
headerFont.bold = True          # Setting column header font to bold
cellFont.bold = False           # Setting cell font to bold

titleFont.height = 240          # 12*20 = 240 Font Size
headerFont.height = 220         # 10*20 = 240 Font Size
cellFont.height = 220           # 10*20 = 240 Font Size

brdLeft = Borders()                # Defining border which is around header
brdLeft.left = 0x01

brdRight = Borders()                # Defining border which is around header
brdRight.right = 0x01

brdTop = Borders()                # Defining border which is around header
brdTop.top = 0x01

brdBottom = Borders()                # Defining border which is around header
brdBottom.bottom = 0x01

brdTopLeft = Borders()
brdTopLeft.top = 0x01
brdTopLeft.left = 0x01

brdBottomLeft = Borders()
brdBottomLeft.bottom = 0x01
brdBottomLeft.left = 0x01

brdBottomRight = Borders()
brdBottomRight.bottom = 0x01
brdBottomRight.right = 0x01

brdTopRight = Borders()
brdTopRight.top = 0x01
brdTopRight.right = 0x01

dateStyle = XFStyle()


titleStyle = XFStyle()
headerStyle = XFStyle()
cellStyle = XFStyle()
dateStyle = XFStyle()

leftCellStyle = XFStyle()
rightCellStyle = XFStyle()
bottomCellStyle = XFStyle()
topleftCellStyle = XFStyle()
bottomleftCellStyle = XFStyle()
bottomrightCellStyle = XFStyle()
toprightCellStyle = XFStyle()

titleStyle.font = titleFont
headerStyle.font = headerFont
headerStyle.borders = brdTop
cellStyle.font = cellFont

topleftCellStyle.font = headerFont
topleftCellStyle.borders = brdTopLeft

bottomleftCellStyle.font = cellFont
bottomleftCellStyle.borders = brdBottomLeft

bottomrightCellStyle.font = cellFont
bottomrightCellStyle.borders = brdBottomRight

toprightCellStyle.font = headerFont
toprightCellStyle.borders = brdTopRight

leftCellStyle.borders = brdLeft
leftCellStyle.font = cellFont

rightCellStyle.borders = brdRight
rightCellStyle.font = cellFont

bottomCellStyle.borders = brdBottom
bottomCellStyle.font = cellFont

pat1 = Pattern()
pat1.pattern = Pattern.SOLID_PATTERN
pat1.pattern_fore_colour = 0x16
headerStyle.pattern = pat1
topleftCellStyle.pattern = pat1
toprightCellStyle.pattern = pat1

def open_data_with_excel(title, headerList, dataList):
    import sys
    import tempfile
    _xls_fd, xls_fn = tempfile.mkstemp(suffix='.xls')
    write_data_to_excel(xls_fn, title, headerList, dataList)
    from PyQt4 import QtGui, QtCore
    if not 'win' in sys.platform:
        QtGui.QDesktopServices.openUrl(QtCore.QUrl('file://%s' % xls_fn))
    else:
        import pythoncom
        import win32com.client
        pythoncom.CoInitialize()
        excel_app = win32com.client.Dispatch("Excel.Application")
        excel_app.Visible = True
        excel_app.Workbooks.Open(xls_fn)

def write_data_to_excel(filename, title, headerList, dataList):
    """
    @param filename: the file to which to save the exported excel
    @param title: title to put in the first row of the genarated excel file
    @param headerList: list of header definitions
    @param dataList: list or generator with the row data
    """
    w = Workbook()
    ws = w.add_sheet('Sheet1')
    ## Writing Title
    ws.write(0, 0, title, titleStyle)                   # Writing Title
    ws.col(0).width = len(title) * 400                  # Setting cell width
    ## Writing Header
    myDataTypeDict = {}            # dictionary of datatype, {columnnumber, Datatype}
    myPrecisionDict = {}        # dictionary of precision , {columnnumber, Precision}
    myLengthDict = {}           # dictionary of length , {columnnumber, length}
    myFormatDict = {}           # dictionary of dateformat , {columnnumber, format}
    for n,desc in enumerate(headerList):
        lst =  desc[1]
        if n==0:
            ws.write(2, n, lst['name'], topleftCellStyle)
        elif n==len(headerList)-1:
            ws.write(2, n, lst['name'], toprightCellStyle)
        else:
            ws.write(2, n, lst['name'], headerStyle)
        if len(lst['name']) < 8:
            ws.col(n).width = 8 *  375
        else:
            ws.col(n).width = len(lst['name']) *  375
        myDataTypeDict[ n ] = lst["python_type"]
        if lst["python_type"] == float:
            myPrecisionDict [ n ] = lst["precision"]    #Populating precision dictionary
        elif lst["python_type"] == datetime.date:
            myFormatDict [ n ] = lst["format"]          #Populating date Format dictionary
        elif lst["python_type"] == str:
            if 'length' in lst:
                myLengthDict [ n ] = lst["length"]          #Populating Column Length dictionary
    ## Writing Data
    row = 3
    column = 0
    valueAddedInSize = 0
    formatStr = '0'
    for dictCounter in dataList:                       # iterating the dataList, having dictionary
        column = 0
        cellStyle.num_format_str = '0'
        for i in range( 0 , len(dictCounter)): #for i in dictCounter:
            valueAddedInSize = 0
            val = dictCounter[i]
            if val != None:
                if not isinstance(val,(str,unicode,int,float,datetime.datetime,datetime.time,datetime.date,
                                       ExcelFormula.Formula) ):
                    val = unicode(val)
                if myDataTypeDict.has_key(column) == True:
                    if myLengthDict.get(column) != None:
                        if len(val) > myLengthDict[ column ]:
                            val = val[0:myLengthDict[ column ]]
                    elif myDataTypeDict[ column ] == str:
                        formatStr = '0'
                    elif myDataTypeDict[ column ] == int:
                        formatStr = '0'
                    elif myDataTypeDict[ column ] == float:
                        formatStr = '0.'
                        for _j in range( 0 , myPrecisionDict[ column ]):
                            formatStr += '0'
                        valueAddedInSize = len(formatStr) # To fit the cell width + 1 (of dot(.))
                    elif myDataTypeDict[ column ] == datetime.date:
                        formatStr = myFormatDict[column]
                        val = datetime.datetime( day = val.day, year = val.year, month = val.month)
                    elif myDataTypeDict[ column ] == bool:
                        formatStr = '0'
                    else:
                        formatStr = '0'
                cellStyle.num_format_str = formatStr
                bottomCellStyle.num_format_str = formatStr
                rightCellStyle.num_format_str = formatStr
                bottomrightCellStyle.num_format_str = formatStr
                leftCellStyle.num_format_str = formatStr
                bottomleftCellStyle.num_format_str = formatStr
            elif val == None:
                val = ' '
            if row - 2  == len(dataList):
                #we re at the bottom row
                if i==0:
                    ws.write(row , column, val , bottomleftCellStyle)
                elif i  == len(dictCounter)-1:
                    ws.write(row , column, val , bottomrightCellStyle)
                else:
                    ws.write(row , column, val , bottomCellStyle)
            else:
                if i==0:
                    ws.write(row , column, val , leftCellStyle)
                elif  i == len(dictCounter)-1:
                    ws.write(row , column, val , rightCellStyle)
                else:
                    ws.write(row , column, val , cellStyle)
            if ws.col(column).width < (len(unicode( val )) )* 300:
                ws.col(column).width = (len(unicode( val )) + valueAddedInSize )* 300
            column = column + 1
        row = row + 1
    w.save(filename)




from camelot.view.model_thread import post
from choiceseditor import ChoicesEditor

class OneToManyChoicesEditor(ChoicesEditor):
  
    def __init__(self, parent, editable=True, target=None, **kwargs):
        ChoicesEditor.__init__(self, parent, editable, **kwargs)
        assert target!=None
        
        def get_choices():
            return [(o, unicode(o)) for o in target.query.all()]
      
        post(get_choices, self.set_choices) 




import datetime

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customeditor import CustomEditor
from camelot.core import constants
from camelot.view.art import Icon
from camelot.core.utils import ugettext_lazy as _

class DateEditor(CustomEditor):
    """Widget for editing date values"""
  
    def __init__(self,
                 parent=None,
                 editable=True,
                 nullable=True,
                 format=constants.camelot_date_format,
                 **kwargs):
        CustomEditor.__init__(self, parent)
        self.format = format
        self.qdateedit = QtGui.QDateEdit()
        self.qdateedit.setAlignment(Qt.AlignRight|Qt.AlignVCenter)
        self.qdateedit.setDisplayFormat(QtCore.QString(format))
    
        special_date_menu = QtGui.QMenu(self)
        special_date_menu.addAction('Today')
        special_date_menu.addAction('Last date')
        special_date = QtGui.QToolButton(self)
        special_date.setIcon(
            Icon('tango/16x16/apps/office-calendar.png').getQIcon())
        special_date.setAutoRaise(True)
        special_date.setToolTip('Special dates')
        special_date.setMenu(special_date_menu)
        special_date.setPopupMode(QtGui.QToolButton.InstantPopup)
        special_date.setFixedHeight(self.get_height())
    
        if not editable:
            special_date.setEnabled(False)
            self.qdateedit.setEnabled(False)
            self.qdateedit.setButtonSymbols(QtGui.QAbstractSpinBox.NoButtons)
      
        if nullable:
            special_date_menu.addAction('Clear')
            self.qdateedit.setSpecialValueText('0/0/0')
        else:
            self.qdateedit.setCalendarPopup(True)
      
        self.hlayout = QtGui.QHBoxLayout()
        
        self.hlayout.addWidget(special_date)
        self.special_date = special_date
        self.hlayout.addWidget(self.qdateedit)
    
        self.hlayout.setContentsMargins(0, 0, 0, 0)
        self.hlayout.setMargin(0)
        self.hlayout.setSpacing(0)
    
        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.hlayout)
    
        self.minimum = datetime.date.min
        self.maximum = datetime.date.max
        self.set_date_range()
        
        self.qdateedit.setFocus(Qt.OtherFocusReason)
    
        self.setFocusProxy(self.qdateedit)
    
        self.connect(self.qdateedit,
                     QtCore.SIGNAL('editingFinished()'),
                     self.editingFinished)
        self.connect(special_date_menu,
                     QtCore.SIGNAL('triggered(QAction*)'),
                     self.setSpecialDate)
    
    # TODO: consider using QDate.toPyDate(), PyQt4.1
    @staticmethod
    def _python_to_qt(value):
        return QtCore.QDate(value.year, value.month, value.day)
    
    # TODO: consider using QDate.toPyDate(), PyQt4.1
    @staticmethod
    def _qt_to_python(value):
        return datetime.date(value.year(), value.month(), value.day())
    
    def editingFinished(self):
        self.emit(QtCore.SIGNAL('editingFinished()'))
        
    def focusOutEvent(self, event):
        self.emit(QtCore.SIGNAL('editingFinished()'))
    
    # TODO: consider using QDate.toPyDate(), PyQt4.1
    def set_date_range(self):
        qdate_min = DateEditor._python_to_qt(self.minimum)
        qdate_max = DateEditor._python_to_qt(self.maximum)
        self.qdateedit.setDateRange(qdate_min, qdate_max)
    
    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        if value:
            newDate = QtCore.QDate()
            newDate.setDate(value.year, value.month, value.day)
            self.qdateedit.setDate(newDate)
        else:
            self.qdateedit.setDate(self.minimumDate())
      
    def get_value(self):
        value = self.qdateedit.date()
        if value == self.minimumDate():
            value = None
        else:
            value = datetime.date(value.year(), value.month(), value.day())    
        return CustomEditor.get_value(self) or value 
      
    def set_enabled(self, editable=True):
        self.qdateedit.setEnabled(editable)
        self.special_date.setEnabled(editable)
  
    def minimumDate(self):
        return self.qdateedit.minimumDate()
    
    def setMinimumDate(self):
        self.qdateedit.setDate(self.minimumDate())
        self.emit(QtCore.SIGNAL('editingFinished()'))
    
    def setSpecialDate(self, action):
        if action.text().compare('Today') == 0:
            self.qdateedit.setDate(QtCore.QDate.currentDate())
        elif action.text().compare('Last date') == 0:
            self.qdateedit.setDate(QtCore.QDate(2400, 12, 31))
        # minimum date is our special value text
        elif action.text().compare('Clear') == 0:
            self.qdateedit.setDate(self.qdateedit.minimumDate())
        self.emit(QtCore.SIGNAL('editingFinished()'))



from camelot.view.art import Icon
from camelot.admin.application_admin import ApplicationAdmin
from camelot.admin.section import Section

class MyApplicationAdmin(ApplicationAdmin):
  
    def get_sections(self):
        from camelot.model.memento import Memento
        from camelot.model.authentication import Person, Organization
        from camelot.model.i18n import Translation
        return [Section('relation',
                        Icon('tango/24x24/apps/system-users.png'),
                        items = [Person, Organization]),
                Section('configuration',
                        Icon('tango/24x24/categories/preferences-system.png'),
                        items = [Memento, Translation])
                ]



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Proxy representing a collection of entities that live in the model thread.

The proxy represents them in the gui thread and provides access to the data
with zero delay.  If the data is not yet present in the proxy, dummy data is
returned and an update signal is emitted when the correct data is available.
"""

import logging
logger = logging.getLogger( 'camelot.view.proxy.collection_proxy' )

import elixir
import datetime
from PyQt4.QtCore import Qt
from PyQt4 import QtGui, QtCore
import sip

from sqlalchemy.orm.session import Session
from camelot.view.art import Icon
from camelot.view.fifo import fifo
from camelot.view.controls import delegates
from camelot.view.remote_signals import get_signal_handler
from camelot.view.model_thread import gui_function, \
                                      model_function, post


class DelayedProxy( object ):
    """A proxy object needs to be constructed within the GUI thread. Construct
    a delayed proxy when the construction of a proxy is needed within the Model
    thread.  On first occasion the delayed proxy will be converted to a real
    proxy within the GUI thread
    """
  
    @model_function
    def __init__( self, *args, **kwargs ):
        self.args = args
        self.kwargs = kwargs
    
    @gui_function
    def __call__( self ):
        return CollectionProxy( *self.args, **self.kwargs )
    
@model_function
def tool_tips_from_object(obj, columns):
  
    data = []
    
    for col in columns:
        tooltip_getter = col[1]['tooltip']
        if tooltip_getter:
            data.append( tooltip_getter(obj) )
        else:
            data.append( None )
            
    return data

@model_function
def background_colors_from_object(obj, columns):
  
    data = []
    
    for col in columns:
        background_color_getter = col[1]['background_color']
        if background_color_getter:
            background_color = background_color_getter(obj)
            data.append( background_color )
        else:
            data.append( None )
            
    return data
  
@model_function
def strip_data_from_object( obj, columns ):
    """For every column in columns, get the corresponding value from the
    object.  Getting a value from an object is time consuming, so using
    this function should be minimized.
    :param obj: the object of which to get data
    :param columns: a list of columns for which to get data
    """
    row_data = []
  
    def create_collection_getter( o, attr ):
        return lambda: getattr( o, attr )
    
    for _i, col in enumerate( columns ):
        field_attributes = col[1]
        if field_attributes['python_type'] == list:
            row_data.append( DelayedProxy( field_attributes['admin'],
                            create_collection_getter( obj, col[0] ),
                            field_attributes['admin'].get_columns ) )
        else:
            row_data.append( getattr( obj, col[0] ) )
    return row_data
  
@model_function
def stripped_data_to_unicode( stripped_data, obj, columns ):
    """Extract for each field in the row data a 'visible' form of 
    data"""
  
    row_data = []
  
    for field_data, ( _field_name, field_attributes ) in zip( stripped_data, columns ):
        unicode_data = u''
        if 'unicode_format' in field_attributes:
            unicode_format = field_attributes['unicode_format']
            if field_data != None:
                unicode_data = unicode_format( field_data )
        elif 'choices' in field_attributes:
            choices = field_attributes['choices']
            if callable(choices):
                for key, value in choices( obj ):
                    if key == field_data:
                        unicode_data = value
                        continue
            else:
                unicode_data = field_data
        elif isinstance( field_data, DelayedProxy ):
            unicode_data = u'...'
        elif isinstance( field_data, list ):
            unicode_data = u'.'.join( [unicode( e ) for e in field_data] )
        elif isinstance( field_data, datetime.datetime ):
            # datetime should come before date since datetime is a subtype of date
            if field_data.year >= 1900:
                unicode_data = field_data.strftime( '%d/%m/%Y %H:%M' )
        elif isinstance( field_data, datetime.date ):
            if field_data.year >= 1900:
                unicode_data = field_data.strftime( '%d/%m/%Y' )
        elif field_data != None:
            unicode_data = unicode( field_data )
        row_data.append( unicode_data )
    
    return row_data
  
from camelot.view.proxy import ValueLoading

class EmptyRowData( object ):
    def __getitem__( self, column ):
        return ValueLoading
        return None
    
empty_row_data = EmptyRowData()

class CollectionProxy( QtCore.QAbstractTableModel ):
    """The CollectionProxy contains a limited copy of the data in the actual
    collection, usable for fast visualisation in a QTableView 
    """
  
    _header_font = QtGui.QApplication.font()
    _header_font_required = QtGui.QApplication.font()
    _header_font_required.setBold( True )
    
    header_icon = Icon( 'tango/16x16/places/folder.png' )

    item_delegate_changed_signal = QtCore.SIGNAL('itemDelegateChanged')
  
    @gui_function
    def __init__( self, admin, collection_getter, columns_getter,
                 max_number_of_rows = 10, edits = None, flush_changes = True ):
        """@param admin: the admin interface for the items in the collection
    
        @param collection_getter: a function that takes no arguments and returns
        the collection that will be visualized. This function will be called inside
        the model thread, to prevent delays when this function causes the database
        to be hit.
    
        @param columns_getter: a function that takes no arguments and returns the
        columns that will be cached in the proxy. This function will be called
        inside the model thread.
        """
        from camelot.view.model_thread import get_model_thread
        self.logger = logging.getLogger(logger.name + '.%s'%id(self))
        self.logger.debug('initialize query table for %s' % (admin.get_verbose_name()))
        QtCore.QAbstractTableModel.__init__(self)
        self.admin = admin
        self.iconSize = QtCore.QSize( QtGui.QFontMetrics( self._header_font_required ).height() - 4, QtGui.QFontMetrics( self._header_font_required ).height() - 4 )
        self.form_icon = QtCore.QVariant( self.header_icon.getQIcon().pixmap( self.iconSize ) )
        self.validator = admin.create_validator( self )
        self.collection_getter = collection_getter
        self.column_count = 0
        self.flush_changes = flush_changes
        self.delegate_manager = None
        self.mt = get_model_thread()
        # Set database connection and load data
        self.rows = 0
        self._columns = []
        self.max_number_of_rows = max_number_of_rows
        self.cache = {Qt.DisplayRole         : fifo( 10 * self.max_number_of_rows ),
                      Qt.EditRole            : fifo( 10 * self.max_number_of_rows ),
                      Qt.ToolTipRole         : fifo( 10 * self.max_number_of_rows ),
                      Qt.BackgroundColorRole : fifo( 10 * self.max_number_of_rows ), }
        # The rows in the table for which a cache refill is under request
        self.rows_under_request = set()
        # The rows that have unflushed changes
        self.unflushed_rows = set()
        # Set edits
        self.edits = edits or []
        self.rsh = get_signal_handler()
        self.rsh.connect( self.rsh,
                         self.rsh.entity_update_signal,
                         self.handleEntityUpdate )
        self.rsh.connect( self.rsh,
                         self.rsh.entity_delete_signal,
                         self.handleEntityDelete )
        self.rsh.connect( self.rsh,
                         self.rsh.entity_create_signal,
                         self.handleEntityCreate )
    
        def get_columns():
            self._columns = columns_getter()
            return self._columns
      
        post( get_columns, self.setColumns )
#    # the initial collection might contain unflushed rows
        post( self.updateUnflushedRows )
#    # in that way the number of rows is requested as well
        post( self.getRowCount, self.setRowCount )
        self.logger.debug( 'initialization finished' )
        
    
    @model_function
    def updateUnflushedRows( self ):
        """Verify all rows to see if some of them should be added to the
        unflushed rows"""
        for i, e in enumerate( self.collection_getter() ):
            if hasattr(e, 'id') and not e.id:
                self.unflushed_rows.add( i )
        
    def hasUnflushedRows( self ):
        """The model has rows that have not been flushed to the database yet,
        because the row is invalid
        """
        has_unflushed_rows = ( len( self.unflushed_rows ) > 0 )
        self.logger.debug( 'hasUnflushed rows : %s' % has_unflushed_rows )
        return has_unflushed_rows
    
    @model_function
    def getRowCount( self ):
        return len( self.collection_getter() )
    
    @gui_function
    def revertRow( self, row ):
        def create_refresh_entity( row ):
    
            @model_function
            def refresh_entity():
                o = self._get_object( row )
                elixir.session.refresh( o )
                return row, o
        
            return refresh_entity
          
        post( create_refresh_entity( row ), self._revert_row )
    
    def _revert_row(self, row_and_entity ):
        row, entity = row_and_entity
        self.handleRowUpdate( row )
        self.rsh.sendEntityUpdate( self, entity )
    
    @gui_function
    def refresh( self ):
        post( self.getRowCount, self._refresh_content )
    
    @gui_function
    def _refresh_content(self, rows ):
        self.cache = {Qt.DisplayRole         : fifo( 10 * self.max_number_of_rows ),
                      Qt.EditRole            : fifo( 10 * self.max_number_of_rows ),
                      Qt.ToolTipRole         : fifo( 10 * self.max_number_of_rows ),
                      Qt.BackgroundColorRole : fifo( 10 * self.max_number_of_rows ),}
        self.setRowCount( rows )
    
    def set_collection_getter( self, collection_getter ):
        self.collection_getter = collection_getter
        self.refresh()
        
    def get_collection_getter( self ):
        return self.collection_getter
    
    def handleRowUpdate( self, row ):
        """Handles the update of a row when this row might be out of date"""
        self.cache[Qt.DisplayRole].delete_by_row( row )
        self.cache[Qt.EditRole].delete_by_row( row )
        self.cache[Qt.ToolTipRole].delete_by_row( row )
        self.cache[Qt.BackgroundColorRole].delete_by_row( row )
        sig = 'dataChanged(const QModelIndex &, const QModelIndex &)'
        self.emit( QtCore.SIGNAL( sig ),
                  self.index( row, 0 ),
                  self.index( row, self.column_count ) )
    
    def handleEntityUpdate( self, sender, entity ):
        """Handles the entity signal, indicating that the model is out of date"""
        self.logger.debug( '%s %s received entity update signal' % \
                     ( self.__class__.__name__, self.admin.get_verbose_name() ) )
        if sender != self:
            row = self.cache[Qt.DisplayRole].delete_by_entity( entity )
            row = self.cache[Qt.EditRole].delete_by_entity( entity )
            row = self.cache[Qt.ToolTipRole].delete_by_entity( entity )
            row = self.cache[Qt.BackgroundColorRole].delete_by_entity( entity )
            if row != None:
                self.logger.debug( 'updated row %i' % row )
                sig = 'dataChanged(const QModelIndex &, const QModelIndex &)'
                if not sip.isdeleted( self ):
                    self.emit( QtCore.SIGNAL( sig ),
                          self.index( row, 0 ),
                          self.index( row, self.column_count ) )
            else:
                self.logger.debug( 'entity not in cache' )
        else:
            self.logger.debug( 'duplicate update' )
      
    def handleEntityDelete( self, sender, entity ):
        """Handles the entity signal, indicating that the model is out of date"""
        self.logger.debug( 'received entity delete signal' )
        if sender != self:
            self.refresh()
      
    def handleEntityCreate( self, sender, entity ):
        """Handles the entity signal, indicating that the model is out of date"""
        self.logger.debug( 'received entity create signal' )
        if sender != self:
            self.refresh()
      
    def setRowCount( self, rows ):
        """Callback method to set the number of rows
        @param rows the new number of rows
        """
        self.rows = rows
        self.emit( QtCore.SIGNAL( 'layoutChanged()' ) )
    
    def getItemDelegate( self ):
        """:return: a DelegateManager for this model, or None if no DelegateManager yet available
        a DelegateManager will be available once the item_delegate_changed signal has been emitted"""
        self.logger.debug( 'getItemDelegate' )
        return self.delegate_manager
    
    def getColumns( self ):
        """@return: the columns as set by the setColumns method"""
        return self._columns
    
    @gui_function
    def setColumns( self, columns ):
        """Callback method to set the columns
    
        @param columns a list with fields to be displayed of the form [('field_name', field_attributes), ...] as
        returned by the getColumns method of the ElixirAdmin class
        """
        self.logger.debug( 'setColumns' )
        self.column_count = len( columns )
        self._columns = columns
    
        delegate_manager = delegates.DelegateManager()
        delegate_manager.set_columns_desc( columns )
    
        # set a delegate for the vertical header
        delegate_manager.insertColumnDelegate( -1, delegates.PlainTextDelegate(parent = delegate_manager) )
        
        for i, c in enumerate( columns ):
            field_name = c[0]
            self.logger.debug( 'creating delegate for %s' % field_name )
            if 'delegate' in c[1]:
                delegate = c[1]['delegate']( parent = delegate_manager, **c[1] )
                delegate_manager.insertColumnDelegate( i, delegate )
                continue
            elif c[1]['python_type'] == str:
                if c[1]['length']:
                    delegate = delegates.PlainTextDelegate( parent = delegate_manager, maxlength = c[1]['length'] )
                    delegate_manager.insertColumnDelegate( i, delegate )
                else:
                    delegate = delegates.TextEditDelegate( parent = delegate_manager, **c[1] )
                    delegate_manager.insertColumnDelegate( i, delegate )
            else:
                delegate = delegates.PlainTextDelegate(parent = delegate_manager)
                delegate_manager.insertColumnDelegate( i, delegate )
        # Only set the delegate manager when it is fully set up
        self.delegate_manager = delegate_manager
        if not sip.isdeleted( self ):
            self.emit( self.item_delegate_changed_signal )
            self.emit( QtCore.SIGNAL( 'layoutChanged()' ) )
      
    def rowCount( self, index = None ):
        return self.rows
    
    def columnCount( self, index = None ):
        return self.column_count
    
    @gui_function
    def headerData( self, section, orientation, role ):
        """In case the columns have not been set yet, don't even try to get
        information out of them
        """
        if orientation == Qt.Horizontal:
            if section >= self.column_count:
                return QtCore.QAbstractTableModel.headerData( self, section, orientation, role )
            c = self.getColumns()[section]
            
            if role == Qt.DisplayRole:
                return QtCore.QVariant( c[1]['name'] )
              
            elif role == Qt.FontRole:
                if ( 'nullable' in c[1] ) and \
                   ( c[1]['nullable'] == False ):
                    return QtCore.QVariant( self._header_font_required )
                else:
                    return QtCore.QVariant( self._header_font )
                  
            elif role == Qt.SizeHintRole:
                option = QtGui.QStyleOptionViewItem()
                if self.delegate_manager:
                    editor_size = self.delegate_manager.sizeHint( option, self.index( 0, section ) )
                else:
                    editor_size = 0
                if 'minimal_column_width' in c[1]:
                    minimal_column_width = QtGui.QFontMetrics( self._header_font ).size( Qt.TextSingleLine, 'A' ).width()*c[1]['minimal_column_width']
                else:
                    minimal_column_width = 0
                editable = True
                if 'editable' in c[1]:
                    editable = c[1]['editable']
                label_size = QtGui.QFontMetrics( self._header_font_required ).size( Qt.TextSingleLine, unicode(c[1]['name']) + ' ' )
                size = max( minimal_column_width, label_size.width() + 10 )
                if editable:
                    size = max( size, editor_size.width() )
                return QtCore.QVariant( QtCore.QSize( size, label_size.height() + 10 ) )
        else:
            if role == Qt.SizeHintRole:
                return QtCore.QVariant( QtCore.QSize( self.iconSize.width() + 8, self.iconSize.height() + 5 ) )
            if role == Qt.DecorationRole:
                return self.form_icon
#      elif role == Qt.DisplayRole:
#        return QtCore.QVariant()
        return QtCore.QAbstractTableModel.headerData( self, section, orientation, role )
    
    @gui_function
    def data( self, index, role ):
        if not index.isValid() or \
           not ( 0 <= index.row() <= self.rowCount( index ) ) or \
           not ( 0 <= index.column() <= self.columnCount( index ) ):
            return QtCore.QVariant()
        if role in ( Qt.DisplayRole, Qt.EditRole, Qt.ToolTipRole,):
            data = self._get_row_data( index.row(), role )
            try:
                value = data[index.column()]
                if isinstance( value, DelayedProxy ):
                    value = value()
                    data[index.column()] = value
                if isinstance( value, datetime.datetime ):
                    # Putting a python datetime into a QVariant and returning it to a PyObject seems
                    # to be buggy, therefor we chop the microseconds
                    if value:
                        value = QtCore.QDateTime(value.year, value.month, value.day, value.hour, value.minute, value.second)
                self.logger.debug( 'get data for row %s;col %s; role %s : %s' % ( index.row(), index.column(), role, unicode( value ) ) )
            except KeyError:
                self.logger.error( 'Programming error, could not find data of column %s in %s' % ( index.column(), str( data ) ) )
                value = None
            return QtCore.QVariant( value )
        elif role == Qt.ForegroundRole:
            pass
        elif role == Qt.BackgroundRole:
            data = self._get_row_data( index.row(), role )
            try:
                value = data[index.column()]
            except:
                self.logger.error( 'Programming error, could not find data of column %s in %s' % ( index.column(), str( data ) ) )
                value = None
            if value in (None, ValueLoading):
                return QtCore.QVariant(QtGui.QColor('white'))
            else:
                return QtCore.QVariant(value)
        return QtCore.QVariant()
    
    def setData( self, index, value, role = Qt.EditRole ):
        """Value should be a function taking no arguments that returns the data to
        be set
        
        This function will then be called in the model_thread
        """
        if role == Qt.EditRole:
    
            flushed = ( index.row() not in self.unflushed_rows )
            self.unflushed_rows.add( index.row() )
      
            def make_update_function( row, column, value ):
      
                @model_function
                def update_model_and_cache():
                    attribute, field_attributes = self.getColumns()[column]
                    # if the field is not editable, don't waste any time and get out of here
                    if not field_attributes['editable']:
                        return False
                      
                    from sqlalchemy.exceptions import DatabaseError
                    from sqlalchemy import orm
                    new_value = value()
                    self.logger.debug( 'set data for row %s;col %s' % ( row, column ) )
          
                    if new_value == ValueLoading:
                        return None
                      
                    o = self._get_object( row )
                    if not o:
                        # the object might have been deleted from the collection while the editor
                        # was still open
                        self.logger.debug( 'this object is no longer in the collection' )
                        try:
                            self.unflushed_rows.remove( row )
                        except KeyError:
                            pass
                        return
                    
                    old_value = getattr( o, attribute )
                    changed = ( new_value != old_value )
                    #
                    # In case the attribute is a OneToMany or ManyToMany, we cannot simply compare the
                    # old and new value to know if the object was changed, so we'll
                    # consider it changed anyway
                    #
                    direction = field_attributes.get( 'direction', None )
                    if direction in ( orm.interfaces.MANYTOMANY, orm.interfaces.ONETOMANY ):
                        changed = True
                    if changed and field_attributes['editable'] == True:
                        # update the model
                        model_updated = False
                        try:
                            setattr( o, attribute, new_value )
                            model_updated = True
                        except AttributeError, e:
                            self.logger.error( u"Can't set attribute %s to %s" % ( attribute, unicode( new_value ) ), exc_info = e )
                        except TypeError:
                            # type error can be raised in case we try to set to a collection
                            pass
                        # update the cache
                        row_data = strip_data_from_object( o, self.getColumns() )
                        self.cache[Qt.EditRole].add_data( row, o, row_data )
                        self.cache[Qt.ToolTipRole].add_data( row, o, tool_tips_from_object( o, self.getColumns()) )
                        self.cache[Qt.BackgroundColorRole].add_data( row, o, background_colors_from_object( o, self.getColumns()) )
                        self.cache[Qt.DisplayRole].add_data( row, o, stripped_data_to_unicode( row_data, o, self.getColumns() ) )
                        if self.flush_changes and self.validator.isValid( row ):
                            # save the state before the update
                            try:
                                elixir.session.flush( [o] )
                            except DatabaseError, e:
                                #@todo: when flushing fails, the object should not be removed from the unflushed rows ??
                                self.logger.error( 'Programming Error, could not flush object', exc_info = e )
                            try:
                                self.unflushed_rows.remove( row )
                            except KeyError:
                                pass
                            #
                            # we can only track history if the model was updated, and it was
                            # flushed before, otherwise it has no primary key yet
                            #
                            if model_updated and hasattr(o, 'id') and o.id:
                                #
                                # in case of images or relations, we cannot pickle them
                                #
                                if ( not 'Imag' in old_value.__class__.__name__ ) and not direction:
                                    from camelot.model.memento import BeforeUpdate
                                    from camelot.model.authentication import getCurrentAuthentication
                                    history = BeforeUpdate( model = unicode( self.admin.entity.__name__ ),
                                                           primary_key = o.id,
                                                           previous_attributes = {attribute:old_value},
                                                           authentication = getCurrentAuthentication() )
                  
                                    try:
                                        elixir.session.flush( [history] )
                                    except DatabaseError, e:
                                        self.logger.error( 'Programming Error, could not flush history', exc_info = e )
                        #@todo: update should only be sent remotely when flush was done 
                        self.rsh.sendEntityUpdate( self, o )
                        return ( ( row, 0 ), ( row, len( self.getColumns() ) ) )
                    elif flushed:
                        self.logger.debug( 'old value equals new value, no need to flush this object' )
                        try:
                            self.unflushed_rows.remove( row )
                        except KeyError:
                            pass
              
                return update_model_and_cache
        
            post( make_update_function( index.row(), index.column(), value ), self._emit_changes )
      
        return True
    
    def _emit_changes( self, region ):
        if region:
            self.emit( QtCore.SIGNAL( 'dataChanged(const QModelIndex &, const QModelIndex &)' ),
                       self.index( region[0][0], region[0][1] ), self.index( region[1][0], region[1][1] ) )
      
    def flags( self, index ):
        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable
        if self.getColumns()[index.column()][1]['editable']:
            flags = flags | Qt.ItemIsEditable
        return flags
    
    @model_function
    def _extend_cache( self, offset, limit ):
        """Extend the cache around row"""
        #@TODO : also store the primary key, here we just saved the id
        columns = self.getColumns()
        offset = min( offset, self.rows )
        limit = min( limit, self.rows - offset )
        for i, o in enumerate( self.collection_getter()[offset:offset + limit + 1] ):
            row_data = strip_data_from_object( o, columns )
            self.cache[Qt.EditRole].add_data( i + offset, o, row_data )
            self.cache[Qt.ToolTipRole].add_data( i + offset, o, tool_tips_from_object( o, self.getColumns()) )
            self.cache[Qt.BackgroundColorRole].add_data( i + offset, o, background_colors_from_object( o, self.getColumns()) )
            self.cache[Qt.DisplayRole].add_data( i + offset, o, stripped_data_to_unicode( row_data, o, columns ) )
        return ( offset, limit )
    
    @model_function
    def _get_object( self, row ):
        """Get the object corresponding to row"""
        try:
            # first try to get the primary key out of the cache, if it's not
            # there, query the collection_getter
            return self.cache[Qt.EditRole].get_entity_at_row( row )
        except KeyError:
            pass
        return self.collection_getter()[row]
    
    def _cache_extended( self, interval ):
        offset, limit = interval
        self.rows_under_request.difference_update( set( range( offset, offset + limit ) ) )
        self.emit( QtCore.SIGNAL( 'dataChanged(const QModelIndex &, const QModelIndex &)' ),
                  self.index( offset, 0 ), self.index( offset + limit, self.column_count ) )
    
    def _get_row_data( self, row, role ):
        """Get the data which is to be visualized at a certain row of the
        table, if needed, post a refill request the cache to get the object
        and its neighbours in the cache, meanwhile, return an empty object
        @param role: Qt.EditRole or Qt.DisplayRole 
        """
        role_cache = self.cache[role]
        try:
            return role_cache.get_data_at_row( row )
        except KeyError:
            if row not in self.rows_under_request:
                offset = max( row - self.max_number_of_rows / 2, 0 )
                limit = self.max_number_of_rows
                self.rows_under_request.update( set( range( offset, offset + limit ) ) )
                post( lambda :self._extend_cache( offset, limit ), self._cache_extended )
            return empty_row_data
      
    @model_function
    def remove( self, o ):
        self.collection_getter().remove( o )
        self.rows -= 1
    
    @model_function
    def append( self, o ):
        self.collection_getter().append( o )
        self.rows += 1
    
    @model_function
    def removeEntityInstance( self, o, delete = True ):
        """Remove the entity instance o from this collection
        @param o: the object to be removed from this collection
        @param delete: delete the object after removing it from the collection 
        """
        self.logger.debug( 'remove entity instance')
        self.remove( o )
        # remove the entity from the cache
        self.cache[Qt.DisplayRole].delete_by_entity( o )
        self.cache[Qt.ToolTipRole].delete_by_entity( o )
        self.cache[Qt.BackgroundColorRole].delete_by_entity( o )
        self.cache[Qt.EditRole].delete_by_entity( o )
        if delete:
            self.rsh.sendEntityDelete( self, o )
            self.admin.delete( o )
        else:
            # even if the object is not deleted, it needs to be flushed to make
            # sure it's out of the collection
            self.admin.flush( o )
        post( self.getRowCount, self._refresh_content )
    
    @gui_function
    def removeRow( self, row, delete = True ):
        """Remove the entity associated with this row from this collection
        @param delete: delete the entity as well
        """
        self.logger.debug( 'remove row %s' % row )
    
        def create_delete_function( row ):
    
            def delete_function():
                o = self._get_object( row )
                self.removeEntityInstance( o, delete )
        
            return delete_function
      
        post( create_delete_function( row ) )
        return True
     
    @gui_function 
    def copy_row( self, row ):
        """Copy the entity associated with this row to the end of the collection
        :param row: the row number
        """
        
        def create_copy_function( row ):
            
            def copy_function():
                o = self._get_object(row)
                new_object = self.admin.entity()
                new_object.from_dict( o.to_dict(exclude=['id']) )
                self.insertEntityInstance(self.getRowCount(), new_object)
                
            return copy_function
                
        post( create_copy_function( row ) )
        return True
    
    @model_function
    def insertEntityInstance( self, row, o ):
        """Insert object o into this collection
        :param o: the object to be added to the collection
        :return: the row at which the object was inserted
        """
        self.append( o )
        row = self.getRowCount() - 1
        self.unflushed_rows.add( row )
        if self.flush_changes and not len( self.validator.objectValidity( o ) ):
            elixir.session.flush( [o] )
            try:
                self.unflushed_rows.remove( row )
            except KeyError:
                pass
# TODO : it's not because an object is added to this list, that it was created
# it might as well exist allready, eg. manytomany relation
#      from camelot.model.memento import Create
#      from camelot.model.authentication import getCurrentAuthentication
#      history = Create(model=unicode(self.admin.entity.__name__),
#                       primary_key=o.id,
#                       authentication = getCurrentAuthentication())
#      elixir.session.flush([history])
#      self.rsh.sendEntityCreate(self, o)
        post( self.getRowCount, self._refresh_content )
        return row
    
    @gui_function
    def insertRow( self, row, entity_instance_getter ):
  
        def create_insert_function( getter ):
    
            @model_function
            def insert_function():
                self.insertEntityInstance( row, getter() )
        
            return insert_function
      
        post( create_insert_function( entity_instance_getter ) )
    
    @model_function
    def getData( self ):
        """Generator for all the data queried by this proxy"""
        for _i, o in enumerate( self.collection_getter() ):
            yield strip_data_from_object( o, self.getColumns() )
            
    def get_admin( self ):
        """Get the admin object associated with this model"""
        return self.admin
      



from camelot.view.controls.exception import model_thread_exception_message_box
from camelot.view.model_thread import post

from PyQt4 import QtGui, QtCore

class OpenFileProgressDialog(QtGui.QProgressDialog):

    def __init__(self):
        QtGui.QProgressDialog.__init__(self, 'Open file', QtCore.QString(), 0, 0)
        self.setRange(0, 0)

    def open_path(self, path):
        url = QtCore.QUrl.fromLocalFile(path)
        QtGui.QDesktopServices.openUrl(url)
        self.close()

class SaveFileProgressDialog(QtGui.QProgressDialog):

    def __init__(self):
        QtGui.QProgressDialog.__init__(self, 'Save file', QtCore.QString(), 0, 0)
        self.setRange(0, 0)

    def finish(self, on_finish):
        on_finish()
        self.close()

def open_stored_file(parent, stored_file):
    """Open the stored file with the default system editor for this file type"""

    progress = OpenFileProgressDialog()

    def get_path():
        return stored_file.storage.checkout(stored_file)

    post(get_path, progress.open_path, model_thread_exception_message_box)
    progress.exec_()

def create_stored_file(parent, storage, on_finish, filter="""All files (*)"""):
    """Popup a QFileDialog, put the selected file in the storage and return the
    call on_finish with the StoredFile when done"""
    filename = QtGui.QFileDialog.getOpenFileName(parent, 'Open file',
                                                 QtCore.QDir.currentPath(),
                                                 filter)
    if filename:
        progress = SaveFileProgressDialog()

        def checkin():
            new_path = storage.checkin(str(filename))
            return lambda:on_finish(new_path)

        post(checkin, progress.finish, model_thread_exception_message_box)
        progress.exec_()



import logging

logger = logging.getLogger( 'camelot.view.controls.editors.onetomanyeditor' )
from customeditor import CustomEditor, QtCore, QtGui, Qt
from wideeditor import WideEditor

from camelot.view.art import Icon
from camelot.view.model_thread import gui_function, model_function, post
from camelot.core.utils import ugettext as _

class One2ManyEditor( CustomEditor, WideEditor ):

    def __init__( self,
                 admin = None,
                 parent = None,
                 create_inline = False,
                 editable = True,
                 **kw ):
        """
    :param admin: the Admin interface for the objects on the one side of the
    relation
    
    :param create_inline: if False, then a new entity will be created within a
    new window, if True, it will be created inline
    
    after creating the editor, setEntityInstance needs to be called to set the
    actual data to the editor
    """
    
        CustomEditor.__init__( self, parent )
        layout = QtGui.QHBoxLayout()
        layout.setContentsMargins( 0, 0, 0, 0 )
        #
        # Setup table
        #
        from camelot.view.controls.tableview import TableWidget
        # parent set by layout manager
        self.table = TableWidget()
        rowHeight = QtGui.QFontMetrics( self.font() ).height() + 5
        self.table.verticalHeader().setDefaultSectionSize( rowHeight )
        layout.setSizeConstraint( QtGui.QLayout.SetNoConstraint )
        self.setSizePolicy( QtGui.QSizePolicy.Expanding,
                           QtGui.QSizePolicy.Expanding )
        self.connect( self.table.verticalHeader(),
                     QtCore.SIGNAL( 'sectionClicked(int)' ),
                     self.createFormForIndex )
        self.admin = admin
        self.editable = editable
        self.create_inline = create_inline
        layout.addWidget( self.table )
        self.setupButtons( layout )
        self.setLayout( layout )
        self.model = None
    
    def setupButtons( self, layout ):
        button_layout = QtGui.QVBoxLayout()
        button_layout.setSpacing( 0 )
        delete_button = QtGui.QToolButton()
        icon = Icon( 'tango/16x16/places/user-trash.png' ).getQIcon()
        delete_button.setIcon( icon )
        delete_button.setAutoRaise( True )
        delete_button.setToolTip(_('delete'))
        self.connect( delete_button,
                     QtCore.SIGNAL( 'clicked()' ),
                     self.deleteSelectedRows )
        add_button = QtGui.QToolButton()
        icon = Icon( 'tango/16x16/actions/document-new.png' ).getQIcon()
        add_button.setIcon( icon )
        add_button.setAutoRaise( True )
        add_button.setToolTip(_('new'))
        self.connect( add_button, QtCore.SIGNAL( 'clicked()' ), self.newRow )
        copy_button = QtGui.QToolButton()
        icon = Icon( 'tango/16x16/actions/edit-copy.png' ).getQIcon()
        copy_button.setIcon( icon )
        copy_button.setAutoRaise( True )
        copy_button.setToolTip(_('copy'))
        self.connect( copy_button, QtCore.SIGNAL( 'clicked()' ), self.copy_selected_rows )        
        export_button = QtGui.QToolButton()
        export_button.setIcon( Icon( 'tango/16x16/mimetypes/x-office-spreadsheet.png' ).getQIcon() )
        export_button.setAutoRaise( True )
        export_button.setToolTip(_('export as spreadsheet'))
        self.connect( export_button,
                     QtCore.SIGNAL( 'clicked()' ),
                     self.exportToExcel )
        button_layout.addStretch()
        if self.editable:
            button_layout.addWidget( add_button )
            button_layout.addWidget( copy_button )
            button_layout.addWidget( delete_button )
        button_layout.addWidget( export_button )
        layout.addLayout( button_layout )
    
    def exportToExcel( self ):
        from camelot.view.export.excel import open_data_with_excel
    
        def export():
            title = self.admin.get_verbose_name_plural()
            columns = self.admin.get_columns()
            if self.model:
                data = list( self.model.getData() )
                open_data_with_excel( title, columns, data )
        
        post( export )
    
    def getModel( self ):
        return self.model
    
    def update_delegates( self, *args ):
        if self.model:
            delegate = self.model.getItemDelegate()
            if delegate:
                self.table.setItemDelegate( delegate )
                for i in range( self.model.columnCount() ):
                    txtwidth = self.model.headerData( i, Qt.Horizontal, Qt.SizeHintRole ).toSize().width()
                    colwidth = self.table.columnWidth( i )
                    self.table.setColumnWidth( i, max( txtwidth, colwidth ) )
          
    def set_value( self, model ):
        model = CustomEditor.set_value( self, model )
        if model:
            self.model = model
            self.table.setModel( model )
      
            def create_fill_model_cache( model ):
              
                def fill_model_cache():
                    model._extend_cache( 0, 10 )
          
                return fill_model_cache
        
            post( create_fill_model_cache( model ), self.update_delegates )
      
    @gui_function
    def activate_editor( self, row ):
        index = self.model.index( row, 0 )
        self.table.scrollToBottom()
        self.table.setCurrentIndex( index )
        self.table.edit( index )
            
    def newRow( self ):
        from camelot.view.workspace import get_workspace
        workspace = get_workspace()
    
        if self.create_inline:
    
            @model_function
            def create():
                o = self.admin.entity()
                row = self.model.insertEntityInstance( 0, o )
                self.admin.set_defaults( o )
                return row
        
            post( create, self.activate_editor )
      
        else:
            prependentity = lambda o: self.model.insertEntityInstance( 0, o )
            removeentity = lambda o: self.model.removeEntityInstance( o )
            #
            # We cannot use the workspace as a parent, in case of working with 
            # the NoDesktopWorkspaces
            #
            form = self.admin.create_new_view( parent = None,
                                               oncreate = prependentity,
                                               onexpunge = removeentity )
            workspace.addSubWindow( form )
            form.show()
            
    def copy_selected_rows( self ):
        """Copy the selected rows in this tableview"""
        logger.debug( 'delete selected rows called' )
        for row in set( map( lambda x: x.row(), self.table.selectedIndexes() ) ):
            self.model.copy_row( row )        
      
    def deleteSelectedRows( self ):
        """Delete the selected rows in this tableview"""
        logger.debug( 'delete selected rows called' )
        for row in set( map( lambda x: x.row(), self.table.selectedIndexes() ) ):
            self.model.removeRow( row )
      
    def createFormForIndex( self, index ):
        from camelot.view.proxy.collection_proxy import CollectionProxy
        from camelot.view.workspace import get_workspace
        model = CollectionProxy( self.admin,
                                self.model.collection_getter,
                                self.admin.get_fields,
                                max_number_of_rows = 1,
                                edits = None )
        form = self.admin.create_form_view( u'', model, index, get_workspace() )
        get_workspace().addSubWindow( form )
        form.show()



from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customeditor import CustomEditor
import re

class CodeEditor(CustomEditor):
  
    def __init__(self, parent=None, parts=['99','AA'], editable=True, **kwargs):
        CustomEditor.__init__(self, parent)
        self.setFocusPolicy(Qt.StrongFocus)
        self.parts = parts
        self.part_editors = []
        layout = QtGui.QHBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(0)
        layout.setAlignment(Qt.AlignLeft)
        for part in parts:
            part = re.sub('\W*', '', part)
            part_length = len(part)
            editor = QtGui.QLineEdit()
            editor.setInputMask(part)
            if not editable:
                editor.setEnabled(False)
            space_width = editor.fontMetrics().size(Qt.TextSingleLine, 'A').width()
            editor.setMaximumWidth(space_width*(part_length+1))
            editor.installEventFilter(self)
            self.part_editors.append(editor)
            layout.addWidget(editor)
            self.connect(editor,
                         QtCore.SIGNAL('editingFinished()'),
                         self.editingFinished)
      
        self.setLayout(layout)
        self.setAutoFillBackground(True)
    
    def editingFinished(self):
        self.emit(QtCore.SIGNAL('editingFinished()'))
        
      
    def set_enabled(self, editable=True):
        for editor in self.part_editors:
            value = editor.text()
            editor.setEnabled(editable)
            editor.setText(value)
      
    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        if value:
            for part_editor, part in zip(self.part_editors, value):
                part_editor.setText(unicode(part))
        else:
            for part_editor in self.part_editors:
                part_editor.setText(u'')
        
    def get_value(self):
        value = []
        for part in self.part_editors:
            value.append(unicode(part.text()))
        return CustomEditor.get_value(self) or value



from PyQt4 import QtGui, QtCore
from PyQt4.QtGui import QWizard, QWizardPage, QToolBar, QFileDialog, QPushButton, QTableView, QFont, QVBoxLayout, QGridLayout, QLabel, QComboBox, QItemDelegate, QStandardItemModel, QColor, QCheckBox
from PyQt4.QtCore import QString, QAbstractTableModel, QVariant, Qt, QAbstractListModel, QModelIndex, QStringList, QPoint
from camelot.view import art
from camelot.view.art import Icon
from camelot.action import createAction, addActions
from camelot.view.elixir_admin import EntityAdmin
from camelot.view.model_thread import get_model_thread
from camelot.view.controls.exception import model_thread_exception_message_box
from camelot.view.controls.delegates.comboboxdelegate import ComboBoxEditorDelegate
from camelot.view.controls.editors.choiceseditor import ChoicesEditor
import csv, itertools

_ = lambda x: x

class ImportWizard(QtGui.QWizard):
    """Import wizard GUI"""
    def __init__(self, parent, attributes):
        QWizard.__init__(self)
        self.parent = parent
        # the attributes of the object that will be imported
        self.attributes = attributes
        print self.attributes

    """ Make a wizard and the pages """
    def start(self):
        self.qWizard = QWizard()
        self.qWizard.setEnabled(True)
        # qPage is the introduction page
        self.qPage = ImportWizardPage(self.qWizard)
        self.qPage.setTitle(QString('import wizard'))

        self.makeToolBarToSearchFile()

        #make grid
        self.makeGridLayout()

        #add layout to page
        self.qPage.setLayout(self.grid)
        self.qWizard.addPage(self.qPage)

        #make the page that shows the table
        self.qTablePage = QWizardPage(self.qWizard)
        self.qTablePage.setTitle(QString('Data from file'))

        self.qWizard.addPage(self.qTablePage)

        cancelButton = QPushButton(QString('cancel'), self.qWizard)
        self.qWizard.setButton(QWizard.CancelButton, cancelButton)

        finishButton = QWizard.FinishButton
        self.qWizard.setButtonText(finishButton, QString('import'))

        self.qWizard.show()
        self.qWizard.exec_()

    """
        makes toolbar with button to open a csv-file. this is added at the qpage of a qwizard
    """
    def makeToolBarToSearchFile(self):
        self.openToolBar = QToolBar(self.qPage)
        icon_file_open = Icon('tango/32x32/actions/fileopen.png').fullpath()
        openAct = QtGui.QAction(QtGui.QIcon(icon_file_open), 'Open File', self.openToolBar)
        self.openToolBar.connect(openAct, QtCore.SIGNAL('triggered()'), self.showOpenFileDialog)
        self.openToolBar.addAction(openAct)

    def makeCheckBoxForFirstRow(self):
        checkBox = QCheckBox('first row of data is column name')
        action = QtGui.QAction('CheckBox', checkBox)
        checkBox.connect(action, QtCore.SIGNAL('clicked()'), self.repaintTable)
        checkBox.addAction(action)
        return checkBox

    def repaintTable(self):
        dataWithoutFirstRow = self.data[1:]
        self.makeTable(dataWithoutFirstRow, self.attributes)


    """
        makes the openfiledialog: when the file is committed, the table is shown.
        the method prepares also the table to show
    """
    def showOpenFileDialog(self):
        filename = QtGui.QFileDialog.getOpenFileName(None, 'Open file', '/')
        #make label
        self.label.clear()
        self.label = QtGui.QLabel(filename, self.qPage)
        self.qPage.initializePath(filename)
        self.grid.addWidget(self.label, 2, 0)

        #open file
        file=open(filename)
        csvreader = csv.reader(file)
        array = list(csvreader)
        self.data = array
        #checkbox
        checkBox = self.makeCheckBoxForFirstRow()
        #tableview
        tableView = self.makeTable(array, self.attributes, False)

        vLayout = QVBoxLayout()
        vLayout.addWidget(checkBox)
        vLayout.addWidget(tableView)
        self.qTablePage.setLayout(vLayout)


    """ the layout for the wizard """
    def makeGridLayout(self):
        self.grid = QtGui.QGridLayout()
        self.grid.setSpacing(10)
        self.grid.addWidget(self.openToolBar, 1, 0)
        self.label = QtGui.QLabel('select file', self.qPage)
        self.grid.addWidget(self.label, 2, 0)

    """ make the table for the page"""
    def makeTable(self, data, headerData, firstRow=False):
        # create the view
        tv = QTableView()

        # set the table model
        tm = InputTableModel(data, headerData, parent=self.qTablePage)

        # add one to the length for the header
        #if firstRow:
        #    tm = QStandardItemModel((len(data) + 2) , len(self.attributes), self.qTablePage)
        #else :
        #    tm = QStandardItemModel((len(data) + 1) , len(self.attributes), self.qTablePage)

        tv.setModel(tm)
        #self.makeHeader(tm, headerData)
        #self.makeBody(tm, data)
        CHOICES = self.makeChoices(headerData)

        delegate = ComboBoxEditorDelegate(choices=lambda o:CHOICES, parent=tv )
        #delegate = TestComboBoxDelegate(self.attributes, parent=tv )
        tv.setItemDelegateForRow(0,delegate)

        # set the minimum size
        self.setMinimumSize(800, 600)

        # hide grid
        tv.setShowGrid(True)

        # set the font
        font = QFont("Courier New", 20)
        tv.setFont(font)

        # hide vertical header
        vh = tv.verticalHeader()
        vh.setVisible(False)

        # hide horizontal header, the first row will be used as header
        hh = tv.horizontalHeader()
        hh.setVisible(False)

        # set column width to fit contents
        # will fale if to much data
        #tv.resizeColumnsToContents()

        # set row height
        nrows = len(list(data))
        for row in xrange(nrows):
            tv.setRowHeight(row, 18)
        return tv

    def makeChoices(self, choices):
        CHOICES = []
        for i in range(len(choices)):
            CHOICES = CHOICES + [(str(i) , choices[i])]
        CHOICES = tuple(CHOICES)


    def makeHeader(self, model, header):
        for column in range(len(header)):
            index = model.index(0, column, QModelIndex())
            model.setData(index, QVariant(header[column]))
            # can be done with the delegate, change this!!!!
            #model.setData(index, QVariant(QColor(Qt.gray)), Qt.BackgroundColorRole)

    def makeBody(self, model, data):
        for row in range(len(data)):
            for column in range(len(self.attributes)):
                index = model.index((row+1), column, QModelIndex())
                model.setData(index, QVariant(self.data[row][column]))

    """method returning the imported data"""
    def getImportedData(self):
        return list(self.data)

class InputTableModel(QAbstractTableModel):
    """ class representing the table """
    def __init__(self, datain, headerData, parent=None, *args):
        QAbstractTableModel.__init__(self, parent, *args)
        # the headerdata will be the first row in the table
        # it is impossible to add a delegate to a qheaderview (it is possible but ignored)
        # so add the data to the first row and add there the delegate
        # you can add a (different) delegate for each row
        #self.headerRow = headerData
        self.fill_up_table(datain, headerData)


    def fill_up_table(self, datain, headerData, hints = None):
        self.arraydata = list(datain)
        self.arraydata.insert(0, headerData)
        self.fill_up_header(hints)

    def fill_up_header(self, hints):
        #first row of data can be a hint
        print "hints" , hints
        if not hints == None :
            self.arraydata.insert(1, hints)
            #for column in range(len(self.arraydata[0])):
            #    self.setData(self.index(0, column), hints[column], Qt.DisplayRole)
        #if no hints are given, fill up the header with first attribute
        #else:
            #for column in range(len(self.arraydata[0])):
            #    self.setData(self.index(0, column), self.arraydata[0][0], Qt.DisplayRole)

        self.emit(QtCore.SIGNAL("layoutChanged()"))


    def rowCount(self, parent):
        #return len(self.arraydata) + len(self.headerRow[0])
        return   len(self.arraydata)

    def columnCount(self, parent):
        return len(self.arraydata[0])

    def data(self, index, role):
        #if headerData != None:
        #     makeHeaderData()
        #TODO, now the table is filled from (row = 0, column = 0)
        # first row will have always a comboboxdelegate (index + 1)
        # second row is maybe a "hint" for the choices of the comboboxdelegate (index + 2)
        # so the actual data starts maybe at the index + 1 or index + 2
        if not index.isValid():
            return QVariant()
        elif role != Qt.DisplayRole:
            return QVariant()
        return QVariant(self.arraydata[index.row()][index.column()])

    def setData(self, index, value, role=Qt.ItemIsEditable):
        self.arraydata[index.row()][index.column()] = value

    """every item is editable, so no need to keep it for each object """
    def flags(self, index):
        return Qt.ItemIsEditable


class ImportWizardPage(QtGui.QWizardPage):
    """
        class for the page shown in the wizard
    """
    def __init__(self, parent=None, path=None, *args):
        QWizardPage.__init__(self, parent, *args)
        self.path = path

    def initializePath(self, path):
        self.path = path
        self.emit(QtCore.SIGNAL('completeChanged()'))

    def isComplete(self):
        return self.path != None



from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import Qt

from customeditor import CustomEditor
from camelot.view.art import Icon

class StarEditor(CustomEditor):

    def __init__(self, parent, maximum=5, editable=True, **kwargs):
        CustomEditor.__init__(self, parent)
        self.setFocusPolicy(Qt.StrongFocus)
        layout = QtGui.QHBoxLayout(self)
        layout.setMargin(0)
        layout.setSpacing(0)
        self.starIcon = Icon('tango/16x16/status/weather-clear.png').getQIcon()
        self.noStarIcon = Icon('tango/16x16/status/weather-clear-noStar.png').getQIcon()
        self.setAutoFillBackground(True)
        #self.starCount = maximum
        self.starCount = 5
        self.buttons = []
        for i in range(self.starCount):
            button = QtGui.QToolButton(self)
            button.setIcon(self.noStarIcon)
            if editable:
                button.setAutoRaise(True)
            else:
                button.setAutoRaise(True)
                button.setDisabled(True)
            button.setFixedHeight(self.get_height())
            self.buttons.append(button)
      
        def createStarClick(i):
            return lambda:self.starClick(i+1)
      
        for i in range(self.starCount):
            self.connect(self.buttons[i],
                         QtCore.SIGNAL('clicked()'),
                         createStarClick(i))
        for i in range(self.starCount):
            layout.addWidget(self.buttons[i])
        layout.addStretch()
        self.setLayout(layout)
    
    def get_value(self):
        return CustomEditor.get_value(self) or self.stars
      
      
    def set_enabled(self, editable=True):
        for button in self.buttons:
            button.setEnabled(editable)
            button.update()
        self.set_value(self.stars)
    
    def starClick(self, value):
        if self.stars == value:
            self.stars -= 1
        else:
            self.stars = int(value)
        for i in range(self.starCount):
            if i+1 <= self.stars:
                self.buttons[i].setIcon(self.starIcon)
            else:
                self.buttons[i].setIcon(self.noStarIcon)
        self.emit(QtCore.SIGNAL('editingFinished()'))
    
    def set_value(self, value):
        value = CustomEditor.set_value(self, value) or 0
        self.stars = int(value)
        for i in range(self.starCount):
            if i+1 <= self.stars:
                self.buttons[i].setIcon(self.starIcon)
            else:
                self.buttons[i].setIcon(self.noStarIcon)



from PyQt4 import QtCore

from camelot.view.controls import editors
from one2manydelegate import One2ManyDelegate

class ManyToManyDelegate(One2ManyDelegate):
    """
  .. image:: ../_static/manytomany.png
  """
    
    def createEditor(self, parent, option, index):
        editor = editors.ManyToManyEditor(parent=parent, **self.kwargs)
        self.setEditorData(editor, index)
        self.connect(editor, 
                     editors.editingFinished,
                     self.commitAndCloseEditor)
        return editor
      
    def commitAndCloseEditor(self):
        editor = self.sender()
        self.emit(QtCore.SIGNAL('commitData(QWidget*)'), editor)
        
    def setModelData(self, editor, model, index):
        if editor.getModel():
            model.setData(index, editor.getModel().collection_getter)



from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import Qt

from customeditor import CustomEditor
from abstractmanytooneeditor import AbstractManyToOneEditor
from camelot.core.utils import variant_to_pyobject, create_constant_function

from camelot.view.art import Icon
from camelot.view.model_thread import gui_function, model_function, post
from camelot.view.search import create_entity_search_query_decorator
from camelot.core.utils import ugettext as _

class Many2OneEditor( CustomEditor, AbstractManyToOneEditor ):
    """Widget for editing many 2 one relations"""
  
    class CompletionsModel( QtCore.QAbstractListModel ):
        def __init__( self, parent = None ):
            QtCore.QAbstractListModel.__init__( self, parent )
            self._completions = []
      
        def setCompletions( self, completions ):
            self._completions = completions
            self.emit( QtCore.SIGNAL( 'layoutChanged()' ) )
      
        def data( self, index, role ):
            if role == Qt.DisplayRole:
                return QtCore.QVariant( self._completions[index.row()][0] )
            elif role == Qt.EditRole:
                return QtCore.QVariant( self._completions[index.row()][1] )
            return QtCore.QVariant()
      
        def rowCount( self, index = None ):
            return len( self._completions )
      
        def columnCount( self, index = None ):
            return 1
      
    def __init__( self, entity_admin = None, parent = None, editable=True, **kwargs ):
        """
    :param entity_admin : The Admin interface for the object on the one side of
    the relation
    """
    
        CustomEditor.__init__( self, parent )
        self.admin = entity_admin
        self._editable = editable
        self.entity_instance_getter = None
        self._entity_representation = ''
        self.entity_set = False
        self.layout = QtGui.QHBoxLayout()
        self.layout.setSpacing( 0 )
        self.layout.setMargin( 0 )
    
        # Search button
        self.search_button = QtGui.QToolButton()
        self.search_button.setFocusPolicy( Qt.ClickFocus )
        self.search_button.setIcon( Icon( 'tango/16x16/actions/edit-clear.png' ).getQIcon() )
        self.search_button.setToolTip(unicode(_('clear')))
        self.search_button.setAutoRaise( True )
        self.search_button.setFixedHeight( self.get_height() )
        self.connect( self.search_button,
                     QtCore.SIGNAL( 'clicked()' ),
                     self.searchButtonClicked )
    
        # Open button
        self.open_button = QtGui.QToolButton()
        self.open_button.setFocusPolicy( Qt.ClickFocus )
        self.open_button.setIcon( Icon( 'tango/16x16/actions/document-new.png' ).getQIcon() )
        self.open_button.setToolTip( unicode(_('new')) )
        self.open_button.setFixedHeight( self.get_height() )
        self.connect( self.open_button,
                     QtCore.SIGNAL( 'clicked()' ),
                     self.openButtonClicked )
        self.open_button.setAutoRaise( True )
    
        # Search input
        self.search_input = QtGui.QLineEdit( self )
        self.setFocusProxy( self.search_input )
        #self.search_input.setReadOnly(True)
        #self.connect(self.search_input, 
        #             QtCore.SIGNAL('returnPressed()'),
        #             self.returnPressed)
        self.connect( self.search_input,
                     QtCore.SIGNAL( 'textEdited(const QString&)' ),
                     self.textEdited )
        # suppose garbage was entered, we need to refresh the content
        self.connect( self.search_input,
                     QtCore.SIGNAL( 'editingFinished()' ),
                     self.editingFinished )
    
        self.completer = QtGui.QCompleter()
        self.completions_model = self.CompletionsModel( self.completer )
        self.completer.setModel( self.completions_model )
        self.completer.setCaseSensitivity( Qt.CaseInsensitive )
        self.completer.setCompletionMode( QtGui.QCompleter.UnfilteredPopupCompletion )
        self.connect( self.completer,
                     QtCore.SIGNAL( 'activated(const QModelIndex&)' ),
                     self.completionActivated )
        self.search_input.setCompleter( self.completer )
    
        # Setup layout
        self.layout.addWidget( self.search_input )
        self.layout.addWidget( self.search_button )
        self.layout.addWidget( self.open_button )
        self.setLayout( self.layout )
        self.set_editable(editable)
    
    def set_editable(self, editable):
        self._editable = editable
        self.search_input.setEnabled(editable)
        self.search_button.setEnabled(editable)
      
    def textEdited( self, text ):
  
        def create_search_completion( text ):
            return lambda: self.search_completions( text )
      
        post( create_search_completion( unicode( text ) ),
              self.display_search_completions )
        self.completer.complete()
    
    @model_function
    def search_completions( self, text ):
        """Search for object that match text, to fill the list of completions
    
    :return: a list of tuples of (object_representation, object_getter)
    """
        search_decorator = create_entity_search_query_decorator( self.admin, text )
        return [( unicode( e ), create_constant_function( e ) )
                for e in search_decorator( self.admin.entity.query ).limit( 20 )]
    
    @gui_function
    def display_search_completions( self, completions ):
        self.completions_model.setCompletions( completions )
        self.completer.complete()
    
    def completionActivated( self, index ):
        object_getter = index.data( Qt.EditRole )
        self.setEntity( variant_to_pyobject(object_getter) )
    
    def openButtonClicked( self ):
        if self.entity_set:
            return self.createFormView()
        else:
            return self.createNew()
      
    def returnPressed( self ):
        if not self.entity_set:
            self.createSelectView()
      
    def searchButtonClicked( self ):
        if self.entity_set:
            self.setEntity( lambda:None )
        else:
            self.createSelectView()
      
    def trashButtonClicked( self ):
        self.setEntity( lambda:None )
    
    @gui_function
    def createNew( self ):
  
        @model_function
        def get_has_subclasses():
            return len( self.admin.get_subclass_tree() )
      
        post( get_has_subclasses, self.show_new_view )
    
    @gui_function
    def show_new_view( self, has_subclasses ):
        selected = QtGui.QDialog.Accepted
        admin = self.admin
        if has_subclasses:
            from camelot.view.controls.inheritance import SubclassDialog
            select_subclass = SubclassDialog( self, self.admin )
            select_subclass.setWindowTitle(_('select'))
            selected = select_subclass.exec_()
            admin = select_subclass.selected_subclass
        if selected:
            from camelot.view.workspace import get_workspace
            workspace = get_workspace()
            form = admin.create_new_view( workspace )
            self.connect( form, form.entity_created_signal, self.selectEntity )
            sub_window = workspace.addSubWindow( form )
            sub_window.show()
      
    def createFormView( self ):
        if self.entity_instance_getter:
    
            def get_admin_and_title():
                object = self.entity_instance_getter()
                admin = self.admin.get_subclass_entity_admin( object.__class__ )
                return admin, ''
        
            post( get_admin_and_title, self.show_form_view)
            
    def show_form_view( self, admin_and_title ):
        admin, title = admin_and_title
    
        def create_collection_getter( instance_getter ):
            return lambda:[instance_getter()]
      
        from camelot.view.proxy.collection_proxy import CollectionProxy
        from camelot.view.workspace import get_workspace
    
        workspace = get_workspace()
        model = CollectionProxy( admin,
                         create_collection_getter( self.entity_instance_getter ),
                         admin.get_fields )
        self.connect( model,
                     QtCore.SIGNAL( 'dataChanged(const QModelIndex &, const QModelIndex &)' ),
                     self.dataChanged )
        form = admin.create_form_view( title, model, 0, workspace )
        workspace.addSubWindow( form )
        form.show()
    
    def dataChanged( self, index1, index2 ):
        self.setEntity( self.entity_instance_getter, False )
    
    def editingFinished( self ):
        self.search_input.setText( self._entity_representation )
    
    def set_value( self, value ):
        value = CustomEditor.set_value( self, value )
        if value:
            self.setEntity( value, propagate = False )
      
    def set_instance_represenation( self, representation_and_propagate ):
        """Update the gui"""
        ((desc, pk), propagate) = representation_and_propagate
        self._entity_representation = desc
        self.search_input.setText( desc )
        if pk != False: 
            self.open_button.setIcon( Icon( 'tango/16x16/places/folder.png' ).getQIcon() )
            self.open_button.setToolTip(unicode(_('open')))
            self.open_button.setEnabled(True)
            self.search_button.setIcon( Icon( 'tango/16x16/actions/edit-clear.png' ).getQIcon() )
            self.search_button.setToolTip(unicode(_('clear')))
            self.entity_set = True
            #self.search_input.setReadOnly(True)
        else:
            self.open_button.setIcon( Icon( 'tango/16x16/actions/document-new.png' ).getQIcon() )
            self.open_button.setToolTip( unicode(_('new')) )
            self.open_button.setEnabled(self._editable)
            self.search_button.setIcon( Icon( 'tango/16x16/actions/system-search.png' ).getQIcon() )
            self.search_button.setToolTip(_('search'))
            self.entity_set = False
            #self.search_input.setReadOnly(False)
        if propagate:
            self.emit( QtCore.SIGNAL( 'editingFinished()' ) )
              
    def setEntity( self, entity_instance_getter, propagate = True ):
  
        def create_instance_getter( entity_instance ):
            return lambda:entity_instance
      
        def get_instance_represenation():
            """Get a representation of the instance
      
      :return: (unicode, pk) its unicode representation and its primary key 
      or ('', False) if the instance was None
      """
            entity = entity_instance_getter()
            self.entity_instance_getter = create_instance_getter( entity )
            if entity and hasattr( entity, 'id' ):
                return (( unicode( entity ), entity.id ), propagate)
            elif entity:
                return (( unicode( entity ), False ), propagate)
            return (( '', False ), propagate)
      
        post( get_instance_represenation, self.set_instance_represenation )
    
    def selectEntity( self, entity_instance_getter ):
        self.setEntity( entity_instance_getter )



"""Decorators to enhance the docstrings of classes
"""

def documented_entity():
    """Class decorator to append an image of the default view for
  an entity to an entity class.  The image can be generated by using
  the testing framework to create images of all default views in an
  application ::

    @documented_entity()
    class Movie(Entity):
      '''A movie as played in the theater'''
      title = Field(Unicode(50))

  The resulting docstring of the Movie entity will be ::

    '''A movie as played in the theater

    image ../_static/entityviews/new_view_movie.png
    '''
  """

    def document_entity(model):
        model.__doc__ = (model.__doc__ or '') + """

    .. image:: ../_static/entityviews/new_view_%s.png
        """%(model.__name__.lower())
        return model

    return document_entity


def documented_type():
    """Class decorator to append an image of the default editor of
  a field type to the docstring of the type"""

    def document_type(field_type):
        field_type.__doc__ = (field_type.__doc__ or '') + """

    .. image:: ../_static/editors/%s_editable.png
    """
        return field_type

    return document_type



from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.core.constants import camelot_small_icon_width
from camelot.core.utils import variant_to_pyobject
from camelot.view.proxy import ValueLoading
from camelot.view.utils import local_date_format

class DateDelegate(CustomDelegate):
    """Custom delegate for date values"""
  
    __metaclass__ = DocumentationMetaclass
    
    editor = editors.DateEditor
    
    def __init__(self, parent=None, editable=True, **kwargs):
        CustomDelegate.__init__(self, parent, editable)
        self.date_format = local_date_format()
        self._width = self._font_metrics.averageCharWidth() * (len(self.date_format) + 4)  + (camelot_small_icon_width*2) * 2 
    
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        
        dateObj = variant_to_pyobject(index.model().data(index, Qt.EditRole))
        
        background_color = QtGui.QColor(index.model().data(index, Qt.BackgroundRole))
        
        if dateObj not in (None, ValueLoading):
            formattedDate = QtCore.QDate(dateObj).toString(self.date_format)
        else:
            formattedDate = "0/0/0"
          
        rect = option.rect
        rect = QtCore.QRect(rect.left()+3, rect.top()+6, 16, 16)
        
        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect(option.rect, option.palette.highlight())
            fontColor = QtGui.QColor()
            if self.editable:
                Color = option.palette.highlightedText().color()
                fontColor.setRgb(Color.red(), Color.green(), Color.blue())
            else:
                fontColor.setRgb(130,130,130)
        else:
            if self.editable:
                painter.fillRect(option.rect, background_color)
                fontColor = QtGui.QColor()
                fontColor.setRgb(0,0,0)
            else:
                painter.fillRect(option.rect, option.palette.window())
                fontColor = QtGui.QColor()
                fontColor.setRgb(130,130,130)
              
              
        painter.setPen(fontColor.toRgb())
        rect = QtCore.QRect(option.rect.left()+23,
                            option.rect.top(),
                            option.rect.width()-23,
                            option.rect.height())
        
        painter.drawText(rect.x()+2,
                         rect.y(),
                         rect.width()-4,
                         rect.height(),
                         Qt.AlignVCenter | Qt.AlignRight,
                         str(formattedDate))
        
        painter.restore()
    
#  def sizeHint(self, option, index):
#    return editors.DateEditor().sizeHint()



"""
Camelot unittest framework
"""

import unittest

_application_ = []

def get_application():
    """Get the singleton QApplication"""
    from PyQt4.QtGui import QApplication
    if not len(_application_):
        import sys
        _application_.append(QApplication(sys.argv))
    return _application_[0]

class ModelThreadTestCase(unittest.TestCase):
    """Base class for implementing test cases that need a running model_thread.
    """

    images_path = ''

    def grab_widget(self, widget, suffix=None, subdir=None):
        """Save a widget as a png file :
    :param widget: the widget to take a screenshot of
    :param suffix: string to add to the default filename of the image
    :param subdir: subdirectory of images_path in which to put the image file, defaults to
    the name of the test class
    - the name of the png file is the name of the test case, without 'test_'
    - it is stored in the directory with the same name as the class, without 'test'
        """
        import sys
        import os
        from PyQt4.QtGui import QPixmap
        if not subdir:
            images_path = os.path.join(self.images_path, self.__class__.__name__.lower()[:-len('Test')])
        else:
            images_path = os.path.join(self.images_path, subdir)
        if not os.path.exists(images_path):
            os.makedirs(images_path)
        test_case_name = sys._getframe(1).f_code.co_name[5:]
        image_name = '%s.png'%test_case_name
        if suffix:
            image_name = '%s_%s.png'%(test_case_name, suffix)
        widget.adjustSize()
        self.process()
        pixmap = QPixmap.grabWidget(widget)
        pixmap = QPixmap.grabWidget(widget)
        pixmap.save(os.path.join(images_path, image_name), 'PNG')

    def process(self):
        """Wait until all events are processed and the queues of the model thread are empty"""
        self.mt.wait_on_work()

    def setUp(self):
        self.app = get_application()
        from camelot.view.model_thread import get_model_thread, construct_model_thread, has_model_thread
        from camelot.view.remote_signals import construct_signal_handler
        if not has_model_thread():
            construct_model_thread()
        construct_signal_handler()
        self.mt = get_model_thread()
        if not self.mt.isRunning():
            self.mt.start()
        # make sure the startup sequence has passed
        self.process()

    def tearDown(self):
        self.process()
        #self.mt.exit(0)
        #self.mt.wait()

class SchemaTest(ModelThreadTestCase):
    """Test the database schema"""

    def test_schema_display(self):
      
        def schema_display_task():
            import os
            from camelot.bin.camelot_manage import schema_display
            schema_display(os.path.join(self.images_path, 'schema.png'))
            
        from camelot.view.model_thread import get_model_thread, post
        post( schema_display_task )
        get_model_thread().wait_on_work()

class EntityViewsTest(ModelThreadTestCase):
    """Test the views of all the Entity subclasses, subclass this class to test all views
    in your application.  This is done by calling the create_table_view and create_new_view
    on a set of admin objects.  To tell the test case which admin objects should be tested,
    overwrite the get_admins method ::

    class MyEntityViewsTest(EntityViewsTest):

        def get_admins(self):
          from elixir import entities
          application_admin import MyApplicationAdmin
          self.app_admin = MyApplicationAdmin()
          return [self.app_admin.get_entity_admin(e) for e in entities if self.app_admin.get_entity_admin(e)]

    """

    def setUp(self):
        super(EntityViewsTest, self).setUp()

    def get_admins(self):
        from elixir import entities
        from camelot.admin.application_admin import ApplicationAdmin
        self.app_admin = ApplicationAdmin()
        return [self.app_admin.get_entity_admin(e) for e in entities if self.app_admin.get_entity_admin(e)]

    def test_table_view(self):
        for admin in self.get_admins():
            widget = admin.create_table_view()
            self.grab_widget(widget, suffix=admin.entity.__name__.lower(), subdir='entityviews')

    def test_new_view(self):
        for admin in self.get_admins():
            widget = admin.create_new_view()
            self.grab_widget(widget, suffix=admin.entity.__name__.lower(), subdir='entityviews')



from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import Qt

from customeditor import CustomEditor
from camelot.view.art import Icon

class SmileyEditor(CustomEditor):

    def __init__(self, parent, img='face-plain', editable=True, **kwargs):
        CustomEditor.__init__(self, parent)
        
        
    
        
        self.box = QtGui.QComboBox()
        
        self.box.setFrame(True)
        self.box.setEditable(False)
        
        
        self.allSmileys = []
        
        self.allSmileys.append('face-angel')
        self.allSmileys.append('face-crying')
        self.allSmileys.append('face-devil-grin')
        self.allSmileys.append('face-glasses')
        self.allSmileys.append('face-grin')
        self.allSmileys.append('face-kiss')
        self.allSmileys.append('face-monkey')
        self.allSmileys.append('face-plain')
        self.allSmileys.append('face-sad')
        self.allSmileys.append('face-smile')
        self.allSmileys.append('face-smile-big')
        self.allSmileys.append('face-surprise')
        self.allSmileys.append('face-wink')
        
        for i, value in enumerate(self.allSmileys):
            imgPath = 'tango/16x16/emotes/' + value + '.png'        
            icon = Icon(imgPath).getQIcon()
            
            self.box.addItem(icon, '')
            self.box.setFixedHeight(self.get_height())
            
            if value == 'face-plain':
                self.box.setCurrentIndex(i)
        
        
        self.setFocusPolicy(Qt.StrongFocus)
        layout = QtGui.QHBoxLayout(self)
        layout.setMargin(0)
        layout.setSpacing(0)
        self.img = img
        self.imgPath = 'tango/16x16/emotes/' + img + '.png'
        self.Icon = Icon(self.imgPath).getQIcon()
        self.setAutoFillBackground(True)
        #self.starCount = maximum
        if not editable:
            self.box.setEnabled(False)
        else:
            self.box.setEnabled(True)
      
      
        self.connect(self.box,
                     QtCore.SIGNAL('currentIndexChanged()'),
                     self.smileyChanged)
    
    
        layout.addWidget(self.box)
        layout.addStretch()
        self.setLayout(layout)
    
    def get_value(self):
        imgIndex = self.box.currentIndex() 
        
        for i, emot in enumerate(self.allSmileys):
            if imgIndex == i:
                imgName = emot
                
        return CustomEditor.get_value(self) or imgName
      
      
    def set_enabled(self, editable=True):
        self.box.setEnabled(editable)
    
    def smileyChanged(self):
      
        value = self.box.currentIndex()  
        
        for i, emot in enumerate(self.allSmileys):
            if value == i:
                imgName = emot
                
        self.emit(QtCore.SIGNAL('editingFinished()'), imgName)
    
    def set_value(self, value):
        value = CustomEditor.set_value(self, value) or 'face-plain'
        self.img = value
        #self.imgPath = 'tango/16x16/emotes/' + self.img + '.png'
        
        for i, smiley in enumerate(self.allSmileys):
            if smiley == self.img:
                self.box.setCurrentIndex(i)
                
        
        



"""Main function, to be called to start the GUI interface"""

def main(application_admin, 
         initialization=lambda:None,
         pre_initialization=lambda:None):
    """Main function, call this function to start the GUI interface
  
    @param application_admin: object of type ApplicationAdmin (as defined in application_admin.py)
    that specifies the look of the GUI interface
    @param initialization: function that will be called during the appearance of the splash
    screen, put all time consuming initialization here.  this function will be called after the
    model thread has been started.
    @param pre-initialization: function that will be called before the model thread has been started,
    but after the QApplication has been created.  This function can be used to run a configuration
    wizard before a connection to the database was made or any gui element has been constructed.
    """
    #
    # before anything else happens or is imported, the splash screen should be there
    #
    import sys
    from PyQt4 import QtGui, QtCore
    app = QtGui.QApplication([a for a in sys.argv if a])
    pixmap = application_admin.get_splashscreen()
    flag = QtCore.Qt.WindowStaysOnTopHint
    splash = QtGui.QSplashScreen(pixmap, flag)
    splash.show()
    
    msgalign = QtCore.Qt.AlignBottom #| QtCore.Qt.AlignRight
    msgcolor = QtCore.Qt.white
  
    # regularly call processEvents to keep the splash alive
    splash.showMessage('Initialize application...', msgalign, msgcolor)
    app.processEvents()
    
#  font = app.font()
#  font.setStyleStrategy(QtGui.QFont.PreferAntialias)
#  font.setPointSize(font.pointSize()+1)
#  app.setFont(font)
    
    import logging
    logger = logging.getLogger('camelot.view.main')
  
    QT_MAJOR_VERSION = float('.'.join(str(QtCore.QT_VERSION_STR).split('.')[0:2]))
    logger.debug('qt version %s, pyqt version %s' % 
                 (QtCore.QT_VERSION_STR, QtCore.PYQT_VERSION_STR))
    logger.debug('qt major version %f' % QT_MAJOR_VERSION)
  
    # regularly call processEvents to keep the splash alive
    app.processEvents()
    
  
    
    import sqlalchemy, elixir
    logger.debug('sqlalchemy version %s'%sqlalchemy.__version__)
    logger.debug('elixir version %s'%elixir.__version__)
  
    # regularly call processEvents to keep the splash alive
    app.processEvents()
      
    app.setOrganizationName(application_admin.get_organization_name())
    app.setOrganizationDomain(application_admin.get_organization_domain())
    app.setApplicationName(application_admin.get_name())
    app.setWindowIcon(application_admin.get_icon())
    pre_initialization()
    app.processEvents()
  
    # regularly call processEvents to keep the splash alive
    splash.showMessage('Setup database...', msgalign, msgcolor)
    app.processEvents()
    #
    # Start the model thread
    #
    from camelot.view.model_thread import get_model_thread, construct_model_thread
    from camelot.view.remote_signals import construct_signal_handler
  
    construct_model_thread()
    construct_signal_handler()
    get_model_thread().start()
    
    #
    # Load camelot translations
    #
    from camelot.core.utils import load_translations
    get_model_thread().post(load_translations)
    splash.showMessage('Load translations...', msgalign, msgcolor)
    translator = QtCore.QTranslator()
    app.installTranslator(translator)
    
    app.processEvents()
    
    # Set the style sheet
    splash.showMessage('Create main window...', msgalign, msgcolor)
    stylesheet = application_admin.get_stylesheet()
    if stylesheet:
        app.setStyleSheet(stylesheet)
      
    # regularly call processEvents to keep the splash alive
    app.processEvents()
    
    # Application specific initialization instructions
    initialization()
  
    # regularly call processEvents to keep the splash alive
    app.processEvents()
    
    mainwindow = application_admin.create_main_window()
      
    mainwindow.show()
    splash.finish(mainwindow)
    sys.exit(app.exec_())



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Tableview"""

import logging
logger = logging.getLogger( 'camelot.view.controls.tableview' )

from PyQt4 import QtCore, QtGui
from PyQt4.QtGui import QSizePolicy
from PyQt4.QtCore import SIGNAL
from PyQt4.QtCore import Qt

from camelot.view.proxy.queryproxy import QueryTableProxy
from camelot.view.controls.view import AbstractView
from camelot.view.model_thread import model_function, gui_function, post

from search import SimpleSearchControl

class TableWidget( QtGui.QTableView):
    """A widget displaying a table, to be used within a TableView"""
  
    def __init__( self, parent = None ):
        QtGui.QTableView.__init__( self, parent )
        logger.debug( 'create TableWidget' )
        self.setSelectionBehavior( QtGui.QAbstractItemView.SelectRows )
        self.setEditTriggers( QtGui.QAbstractItemView.SelectedClicked | QtGui.QAbstractItemView.DoubleClicked )
        self.setSizePolicy( QSizePolicy.Expanding, QSizePolicy.Expanding )
        self.horizontalHeader().setClickable( False )
        self._header_font_required = QtGui.QApplication.font()
        self._header_font_required.setBold( True )
        self._minimal_row_height = QtGui.QFontMetrics(QtGui.QApplication.font()).lineSpacing() + 10
        self.verticalHeader().setDefaultSectionSize( self._minimal_row_height )
    
    def setModel( self, model ):
        QtGui.QTableView.setModel( self, model )
        self.connect( self.selectionModel(), SIGNAL( 'currentChanged(const QModelIndex&,const QModelIndex&)' ), self.activated )
    
    def activated( self, selectedIndex, previousSelectedIndex ):
        option = QtGui.QStyleOptionViewItem()
        newSize = self.itemDelegate( selectedIndex ).sizeHint( option, selectedIndex )
        row = selectedIndex.row()
        if previousSelectedIndex.row() >= 0:
            oldSize = self.itemDelegate( previousSelectedIndex ).sizeHint( option, selectedIndex )
            previousRow = previousSelectedIndex.row()
            self.setRowHeight( previousRow, oldSize.height() )
        self.setRowHeight( row, newSize.height() )
    
class RowsWidget( QtGui.QLabel ):
    """Widget that is part of the header widget, displaying the number of rows
    in the table view"""
  
    _number_of_rows_font = QtGui.QApplication.font()
  
    def __init__( self, parent ):
        QtGui.QLabel.__init__( self, parent )
        self.setFont( self._number_of_rows_font )
    
    def setNumberOfRows( self, rows ):
        self.setText( u'(%i rows)' % rows )
    
class HeaderWidget( QtGui.QWidget ):
    """HeaderWidget for a tableview, containing the title, the search widget,
    and the number of rows in the table"""
  
    search_widget = SimpleSearchControl
    rows_widget = RowsWidget
  
    _title_font = QtGui.QApplication.font()
    _title_font.setBold( True )
  
    def __init__( self, parent, admin ):
        QtGui.QWidget.__init__( self, parent )
        widget_layout = QtGui.QHBoxLayout()
        self.search = self.search_widget( self )
        title = QtGui.QLabel( admin.get_verbose_name_plural().capitalize(), self )
        title.setFont( self._title_font )
        widget_layout.addWidget( title )
        widget_layout.addWidget( self.search )
        if self.rows_widget:
            self.number_of_rows = self.rows_widget( self )
            widget_layout.addWidget( self.number_of_rows )
      
        else:
            self.number_of_rows = None
        self.setLayout( widget_layout )
        self.setSizePolicy( QSizePolicy.Minimum, QSizePolicy.Fixed )
        self.setNumberOfRows( 0 )
    
    @gui_function
    def setNumberOfRows( self, rows ):
        if self.number_of_rows:
            self.number_of_rows.setNumberOfRows( rows )
      
class TableView( AbstractView  ):
    """A generic tableview widget that puts together some other widgets.  The behaviour of this class and
  the resulting interface can be tuned by specifying specific class attributes which define the underlying
  widgets used ::
  
    class MovieRentalTableView(TableView):
      title_format = 'Grand overview of recent movie rentals'
  
  The attributes that can be specified are :
  
  .. attribute:: header_widget
  
  The widget class to be used as a header in the table view::
   
    header_widget = HeaderWidget
    
  .. attribute:: table_widget
  
  The widget class used to display a table within the table view ::
  
  table_widget = TableWidget
  
  .. attribute:: title_format
  
  A string used to format the title of the view ::
  
  title_format = '%(verbose_name_plural)s'
  
  .. attribute:: table_model
  
  A class implementing QAbstractTableModel that will be used as a model for the table view ::
  
    table_model = QueryTableProxy
  
  - emits the row_selected signal when a row has been selected
  """
  
    header_widget = HeaderWidget
    table_widget = TableWidget
  
    #
    # The proxy class to use 
    #
    table_model = QueryTableProxy
    #
    # Format to use as the window title
    #
    title_format = '%(verbose_name_plural)s'
  
    def __init__( self, admin, search_text = None, parent = None ):
        AbstractView.__init__( self, parent )
        self.admin = admin
        post( self.get_title, self.change_title )
        widget_layout = QtGui.QVBoxLayout()
        if self.header_widget:
            self.header = self.header_widget( self, admin )
            widget_layout.addWidget( self.header )
            self.connect( self.header.search, SIGNAL( 'search' ), self.startSearch )
            self.connect( self.header.search, SIGNAL( 'cancel' ), self.cancelSearch )
            if search_text:
                self.header.search.search( search_text )
        else:
            self.header = None
        widget_layout.setSpacing( 0 )
        widget_layout.setMargin( 0 )
        self.splitter = QtGui.QSplitter( self )
        widget_layout.addWidget( self.splitter )
        table_widget = QtGui.QWidget( self )
        filters_widget = QtGui.QWidget( self )
        self.table_layout = QtGui.QVBoxLayout()
        self.table_layout.setSpacing( 0 )
        self.table_layout.setMargin( 0 )
        self.table = None
        self.filters_layout = QtGui.QVBoxLayout()
        self.filters_layout.setSpacing( 0 )
        self.filters_layout.setMargin( 0 )     
        self.filters = None
        self._table_model = None
        table_widget.setLayout( self.table_layout )
        filters_widget.setLayout( self.filters_layout )
        #filters_widget.hide()
        self.set_admin( admin )
        self.splitter.addWidget( table_widget )
        self.splitter.addWidget( filters_widget )
        self.setLayout( widget_layout )
        self.closeAfterValidation = QtCore.SIGNAL( 'closeAfterValidation()' )
        self.search_filter = lambda q: q
        post( self.admin.get_subclass_tree, self.setSubclassTree )
    
    @model_function
    def get_title( self ):
        return self.title_format % {'verbose_name_plural':self.admin.get_verbose_name_plural()}
    
    @gui_function
    def setSubclassTree( self, subclasses ):
        if len( subclasses ) > 0:
            from inheritance import SubclassTree
            class_tree = SubclassTree( self.admin, self.splitter )
            self.splitter.insertWidget( 0, class_tree )
            self.connect( class_tree, SIGNAL( 'subclassClicked' ), self.set_admin )
      
    def sectionClicked( self, section ):
        """emits a row_selected signal"""
        self.emit( SIGNAL( 'row_selected' ), section )
    
    def create_table_model( self, admin ):
        """Create a table model for the given admin interface"""
        return self.table_model( admin,
                                 admin.get_query,
                                 admin.get_columns )
    
    @gui_function
    def set_admin( self, admin ):
        """Switch to a different subclass, where admin is the admin object of the
        subclass"""
        logger.debug('set_admin called')
        self.admin = admin
        if self.table:
            self.disconnect(self._table_model, QtCore.SIGNAL( 'layoutChanged()' ), self.tableLayoutChanged )
            self.table_layout.removeWidget(self.table)
            self.table.deleteLater()
            self._table_model.deleteLater()
        self.table = self.table_widget( self.splitter )
        self._table_model = self.create_table_model( admin )
        self.table.setModel( self._table_model )
        self.connect( self.table.verticalHeader(),
                      SIGNAL( 'sectionClicked(int)' ),
                      self.sectionClicked )
        self.connect( self._table_model, QtCore.SIGNAL( 'layoutChanged()' ), self.tableLayoutChanged )
        self.tableLayoutChanged()
        self.table_layout.insertWidget( 1, self.table )
    
        def get_filters_and_actions():
            return ( admin.get_filters(), admin.get_list_actions() )
      
        post( get_filters_and_actions,  self.set_filters_and_actions )
        post( admin.get_list_charts, self.setCharts )
    
    @gui_function
    def tableLayoutChanged( self ):
        logger.debug('tableLayoutChanged')
        if self.header:
            self.header.setNumberOfRows( self._table_model.rowCount() )
        item_delegate = self._table_model.getItemDelegate()
        if item_delegate:
            self.table.setItemDelegate( item_delegate )
        for i in range( self._table_model.columnCount() ):
            self.table.setColumnWidth( i, max( self._table_model.headerData( i, Qt.Horizontal, Qt.SizeHintRole ).toSize().width(), 
                                               self.table.columnWidth( i ) ) )
      
    @gui_function
    def setCharts( self, charts ):
        """creates and display charts"""
        pass
#    if charts:
#
#      from matplotlib.figure import Figure
#      from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as \
#                                                     FigureCanvas
#
#      chart = charts[0]
#
#      def getData():
#        """fetches data for chart"""
#        from sqlalchemy.sql import select, func
#        from elixir import session
#        xcol = getattr( self.admin.entity, chart['x'] )
#        ycol = getattr( self.admin.entity, chart['y'] )
#        session.bind = self.admin.entity.table.metadata.bind
#        result = session.execute( select( [xcol, func.sum( ycol )] ).group_by( xcol ) )
#        summary = result.fetchall()
#        return [s[0] for s in summary], [s[1] for s in summary]
#
#      class MyMplCanvas( FigureCanvas ):
#        """Ultimately, this is a QWidget (as well as a FigureCanvasAgg)"""
#
#        def __init__( self, parent = None, width = 5, height = 4, dpi = 100 ):
#          fig = Figure( figsize = ( width, height ), dpi = dpi, facecolor = 'w' )
#          self.axes = fig.add_subplot( 111, axisbg = 'w' )
#          # We want the axes cleared every time plot() is called
#          self.axes.hold( False )
#          self.compute_initial_figure()
#          FigureCanvas.__init__( self, fig )
#          self.setParent( parent )
#          FigureCanvas.setSizePolicy( self,
#                                     QSizePolicy.Expanding,
#                                     QSizePolicy.Expanding )
#          FigureCanvas.updateGeometry( self )
#
#
#        def compute_initial_figure( self ):
#          pass
#
#      def setData( data ):
#        """set chart data"""
#
#        class MyStaticMplCanvas( MyMplCanvas ):
#          """simple canvas with a sine plot"""
#
#          def compute_initial_figure( self ):
#            """computes initial figure"""
#            x, y = data
#            bar_positions = [i - 0.25 for i in range( 1, len( x ) + 1 )]
#            width = 0.5
#            self.axes.bar( bar_positions, y, width, color = 'b' )
#            self.axes.set_xlabel( 'Year' )
#            self.axes.set_ylabel( 'Sales' )
#            self.axes.set_xticks( range( len( x ) + 1 ) )
#            self.axes.set_xticklabels( [''] + [str( d ) for d in x] )
#
#        sc = MyStaticMplCanvas( self, width = 5, height = 4, dpi = 100 )
#        self.table_layout.addWidget( sc )
#
#      self.admin.mt.post( getData, setData )

    def deleteSelectedRows( self ):
        """delete the selected rows in this tableview"""
        logger.debug( 'delete selected rows called' )
        for row in set( map( lambda x: x.row(), self.table.selectedIndexes() ) ):
            self._table_model.removeRow( row )
      
    @gui_function
    def newRow( self ):
        """Create a new row in the tableview"""
        from camelot.view.workspace import get_workspace
        workspace = get_workspace()
        form = self.admin.create_new_view( workspace,
                                           oncreate = lambda o:self._table_model.insertEntityInstance( 0, o ),
                                           onexpunge = lambda o:self._table_model.removeEntityInstance( o ) )
        workspace.addSubWindow( form )
        form.show()
    
    def selectTableRow( self, row ):
        """selects the specified row"""
        self.table.selectRow( row )
    
    def selectedTableIndexes( self ):
        """returns a list of selected rows indexes"""
        return self.table.selectedIndexes()
    
    def getColumns( self ):
        """return the columns to be displayed in the table view"""
        return self.admin.get_columns()
    
    def getData( self ):
        """generator for data queried by table model"""
        for d in self._table_model.getData():
            yield d
      
    def getTitle( self ):
        """return the name of the entity managed by the admin attribute"""
        return self.admin.get_verbose_name()
    
    def viewFirst( self ):
        """selects first row"""
        self.selectTableRow( 0 )
    
    def viewLast( self ):
        """selects last row"""
        self.selectTableRow( self._table_model.rowCount() - 1 )
    
    def viewNext( self ):
        """selects next row"""
        first = self.selectedTableIndexes()[0]
        next = ( first.row() + 1 ) % self._table_model.rowCount()
        self.selectTableRow( next )
    
    def viewPrevious( self ):
        """selects previous row"""
        first = self.selectedTableIndexes()[0]
        prev = ( first.row() - 1 ) % self._table_model.rowCount()
        self.selectTableRow( prev )
    
    def rebuildQuery( self ):
        """resets the table model query"""
    
        def rebuild_query():
            query = self.admin.entity.query
            if self.filters:
                query = self.filters.decorate_query( query )
            if self.search_filter:
                query = self.search_filter( query )
            query_getter = lambda:query
            return query_getter
      
        post( rebuild_query, self._table_model.setQuery )
    
    def startSearch( self, text ):
        """rebuilds query based on filtering text"""
        from camelot.view.search import create_entity_search_query_decorator
        logger.debug( 'search %s' % text )
        self.search_filter = create_entity_search_query_decorator( self.admin, text )
        self.rebuildQuery()
    
    def cancelSearch( self ):
        """resets search filtering to default"""
        logger.debug( 'cancel search' )
        self.search_filter = lambda q: q
        self.rebuildQuery()
        
    @gui_function
    def set_filters_and_actions( self, filters_and_actions ):
        """sets filters for the tableview"""
        filters, actions = filters_and_actions
        from filterlist import FilterList
        from actionsbox import ActionsBox
        logger.debug( 'setting filters for tableview' )
        
        if self.filters:
            self.disconnect( self.filters, SIGNAL( 'filters_changed' ), self.rebuildQuery )
            self.filters_layout.removeWidget(self.table)
            self.filters.deleteLater()
            self.filters = None
        if filters:
            self.filters = FilterList( filters, parent=self.splitter )
            self.filters_layout.addWidget( self.filters )
            self.connect( self.filters, SIGNAL( 'filters_changed' ), self.rebuildQuery )
        elif actions:
            self.filters = ActionsBox( self, self._table_model.collection_getter, lambda:[] )
            self.filters.setActions( actions )
            self.filters_layout.addWidget( self.filters )
      
    def toHtml( self ):
        """generates html of the table"""
        table = [[getattr( row, col[0] ) for col in self.admin.get_columns()]
                 for row in self.admin.entity.query.all()]
        context = {
          'title': self.admin.get_verbose_name_plural(),
          'table': table,
          'columns': [field_attributes['name'] for _field, field_attributes in self.admin.get_columns()],
        }
        from camelot.view.templates import loader
        from jinja import Environment
        env = Environment( loader = loader )
        tp = env.get_template( 'table_view.html' )
        return tp.render( context )
    
    def closeEvent( self, event ):
        """reimplements close event"""
        logger.debug( 'tableview closed' )
        # remove all references we hold, to enable proper garbage collection
        del self.table_layout
        del self.table
        del self.filters
        del self._table_model
        event.accept()
    
    def importWizard(self, attributes):
        from camelot.view.wizard.import_data import ImportWizard
        object_attributes = ['title', 'releasedate', 'name', 'description' ]
        importWizard = ImportWizard( self, object_attributes )
        importWizard.start()
        data = importWizard.getImportedData()    
        
        def makeImport():
            for row in data:
                # get all possible fields (=attributes) from this object
                #attributes = o.Admin.form_display.get_fields() 
                #object_attributes = ['title', 'releasedate', 'name', 'description' ]           
                #print attributes
                # set title
                o = self.admin.entity()
                #For example, setattr(x, 'foobar', 123) is equivalent to x.foobar = 123
                setattr(o, object_attributes[0], row[0])
                #movie.title = row[0]
                name = row[2].split( ' ' ) #director
                #director = Person()
                #director.first_name = name[0]  
                #director.last_name = name[1]
                #movie.director = director
                o.short_description = "korte beschrijving"
                #date = row[1].split('/') # date 12/03/2009
                #o.releasedate = datetime.date(year=int(date[2]), month=int(date[1]), day=int(date[0]))
                o.genre = ""
                #print o[attributes[0]]
                from sqlalchemy.orm.session import Session
                Session.object_session(o).flush([o])
    
        post( makeImport )
            
    def importFromFile( self ):
        """"import data : the data will be imported in the activeMdiChild """
        logger.info( 'call import method' )
        from camelot.view.wizard.importwizard import ImportWizard
        wizard = ImportWizard(self, self.admin)
        wizard.exec_()



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================


from camelot.view.model_thread import model_function


_application_admin_ = []


def get_application_admin():
    if not len(_application_admin_):
        raise Exception('No application admin class has been constructed yet')
    return _application_admin_[0]


class ApplicationAdmin(object):
    """The Application Admin class defines how the application should look
    like, it also ties python classes to their associated admin classes.  It's
    behaviour can be steered by overwriting its static attributes or it's
    methods :

    .. attribute:: name

    The name of the application, as it will appear in the title of the main
    window.

    .. attribute:: sections

    A list containing the various sections that should appear in the left panel
    of the mainwindow.

    .. image:: ../_static/picture2.png

    """

    name = 'Camelot'
    sections = ['Relations', 'Configuration']
    admins = {}

    def __init__(self):
        _application_admin_.append(self)

    def register(self, entity, admin_class):
        self.admins[entity] = admin_class

    @model_function
    def get_sections(self):
        from camelot.admin.section import structure_to_sections
        return structure_to_sections(self.sections)

    def get_entity_admin(self, entity):
        """Get the default entity admin for this entity, return None, if not
        existant"""
        try:
            return self.admins[entity](self, entity)
        except KeyError:
            pass
        if hasattr(entity, 'Admin'):
            return entity.Admin(self, entity)

    def get_entity_query(self, entity):
        """Get the root query for an entity"""
        return entity.query

    def create_main_window(self):
        """create_main_window"""
        from camelot.view.mainwindow import MainWindow
        mainwindow = MainWindow(self)
        return mainwindow

    def get_entities_and_queries_in_section(self, section):
        """:return: a list of tuples of (admin, query) instances related to
        the entities in this section.
        """
        result = [(self.get_entity_admin(e), self.get_entity_query(e))
                  for e, a in self.admins.items()
                  if hasattr(a, 'section')
                  and a.section == section]
        result.sort(cmp = lambda x, y: cmp(x[0].get_verbose_name_plural(),
                                           y[0].get_verbose_name_plural()))
        return result

    def get_actions(self):
        """:return: a list of actions that should be added to the menu and the
        icon bar for this application, each action is a tuple of
        (name, icon, callable), where callable is a function taking no
        arguments that will be called when the action is executed.  Callable
        will be called in the model thread.
        """
        return []

    def get_name(self):
        """:return: the name of the application"""
        return self.name

    def get_icon(self):
        """:return: the QIcon that should be used for the application"""
        from camelot.view.art import Icon
        return Icon('tango/32x32/apps/system-users.png').getQIcon()

    def get_splashscreen(self):
        """:return: a QtGui.QPixmap to be used as splash screen"""
        from camelot.view.art import Pixmap
        return Pixmap('splashscreen.png').getQPixmap()

    def get_organization_name(self):
        return 'Conceptive Engineering'

    def get_organization_domain(self):
        return 'conceptive.be'

    def get_help_url(self):
        """:return: a QUrl pointing to the index page for help"""
        from PyQt4.QtCore import QUrl
        return QUrl('http://www.conceptive.be/projects/camelot/')

    def get_stylesheet(self):
        """
        :return: the qt stylesheet to be used for this application as a string
        or None if no stylesheet needed
        """
        return None

    def get_about(self):
        """:return: the content of the About dialog, a string with html
        syntax"""
        return """<b>Camelot Project</b>
                  <p>
                  Copyright &copy; 2008-2009 Conceptive Engineering.
                  All rights reserved.
                  </p>
                  <p>
                  http://www.conceptive.be/projects/camelot
                  </p>
                  """

    def get_default_field_attributes(self, type_, field):
        """Returns the default field attributes"""
        from camelot.core.view.field_attributes import \
            _sqlalchemy_to_python_type_
        return _sqlalchemy_to_python_type_[type_](field)



#  ==================================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ==================================================================================

"""
Classes to connect the QT event loop with a messaging
server.  To enable multiple clients to push model updates
to each other or messages for the users.

As a messaging server, Apache active MQ was tested in combination
with the stomp library (http://docs.codehaus.org/display/STOMP/Python)
"""

import logging
import re

logger = logging.getLogger('remote_signals')

from PyQt4.QtCore import *

class SignalHandler(QObject):
    def __init__(self):
        QObject.__init__(self)
        import settings
        self.entity_update_signal = SIGNAL("entity_update")
        self.entity_delete_signal = SIGNAL("entity_delete")
        self.entity_create_signal = SIGNAL("entity_create")
        self.update_expression = re.compile(r'^/topic/Camelot.Entity.(?P<entity>.*).update$')
        if hasattr(settings, 'CAMELOT_SERVER') and settings.CAMELOT_SERVER:
            from stomp import stomp
            self.connection = stomp.Connection(host_and_ports = [ (settings.CAMELOT_SERVER, 61613) ])
            self.connection.add_listener(self)
            self.connection.start()
            logger.debug('connection to servers started')
        else:
            self.connection = None
            logger.debug('not connected to a server')
    def on_error(self, headers, message):
        logger.error('received an error %s'%message)
    def on_message(self, headers, message):
        from elixir import entities
        logger.debug('received a message %s : %s'%(str(headers),message))
        match = self.update_expression.match(headers['destination'])
        if match:
            entity = match.group('entity')
            logger.debug(' decoded as update signal for entity %s'%entity)
            self.emit(self.entity_update_signal, self, [e for e in entities if e.__name__==entity][0].get(eval(message)))
    def on_connecting(self, server):
        logger.debug('try to connect to message service')
        self.connection.connect()
    def on_connected(self, *args, **kwargs):
        logger.debug('connected to message service %s, %s'%((str(args), str(kwargs))))
        self.connection.subscribe(destination='/topic/Camelot.Entity.>', ack='auto')
    def on_disconnected(self):
        logger.debug('stomp service disconnected')
    def send_entity_update(self, sender, entity, scope='local'):
        self.sendEntityUpdate(sender, entity, scope)    
    def sendEntityUpdate(self, sender, entity, scope='local'):
        # deprecated
        self.emit(self.entity_update_signal, sender, entity)
        if self.connection and scope=='remote':
            self.connection.send(str([entity.id]), destination='/topic/Camelot.Entity.%s.update'%entity.__class__.__name__)
    def sendEntityDelete(self, sender, entity, scope='local'):
        if self.connection and scope=='remote':
            self.connection.send(str([entity.id]), destination='/topic/Camelot.Entity.%s.delete'%entity.__class__.__name__)
    def sendEntityCreate(self, sender, entity, scope='local'):
        if self.connection and scope=='remote':
            self.connection.send(str([entity.id]), destination='/topic/Camelot.Entity.%s.create'%entity.__class__.__name__)

_signal_handler_ = []

def construct_signal_handler(*args, **kwargs):
    _signal_handler_.append(SignalHandler(*args, **kwargs))

def get_signal_handler():
    return _signal_handler_[-1]



class WideEditor(object):
    """Class signaling that an editor, is a wide editor, so it's label should be displayed
  on top of the editor and the editor itself should take two columns::

    class WideTextLineEditor(TextLineEditor, WideEditor):
      pass

  will generate a test line editor where the text line takes the whole with of the
  form"""



from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.view.art import Icon

class SmileyDelegate(CustomDelegate):
    """Delegate for Smiley's
  """
  
    __metaclass__ = DocumentationMetaclass
    
    editor = editors.SmileyEditor
  
    def __init__(self, parent, editable=True,  **kwargs):
        CustomDelegate.__init__(self,
                                parent=parent,
                                editable=editable,
                                maximum=1,
                                **kwargs)
        maximum = 1
        
        self.maximum = maximum
        
        
    
        
    def paint(self, painter, option, index):
        painter.save()
        img = index.model().data(index, Qt.DisplayRole).toString()
        
        background_color = QtGui.QColor(index.model().data(index, Qt.BackgroundRole))
        
        imgPath = 'tango/16x16/emotes/' + img + '.png'
        self.drawBackground(painter, option, index)
        rect = option.rect
        rect = QtCore.QRect(rect.left()+3, rect.top()+6, rect.width()-5, rect.height())
        
        if( option.state & QtGui.QStyle.State_Selected ):
            painter.fillRect(option.rect, option.palette.highlight())
        else:
            if not self.editable:
                painter.fillRect(option.rect, option.palette.window())
            else:
                painter.fillRect(option.rect, background_color)
          
        icon = Icon(imgPath).getQPixmap()
        QtGui.QApplication.style().drawItemPixmap(painter, rect, 1, icon)
        rect = QtCore.QRect(rect.left()+20, rect.top(), rect.width(), rect.height())
        painter.restore()



from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.view.proxy import ValueLoading
from camelot.view.art import Icon
from camelot.core.utils import variant_to_pyobject

class VirtualAddressDelegate(CustomDelegate):
    """
  """
  
    __metaclass__ = DocumentationMetaclass
  
    editor = editors.VirtualAddressEditor
  
    def paint(self, painter, option, index):
        painter.save()
        self.drawBackground(painter, option, index)
        virtual_address = variant_to_pyobject(index.model().data(index, Qt.EditRole))  
  
        if virtual_address and virtual_address!=ValueLoading:
            self.paint_text(painter, option, index, unicode(virtual_address[1] or ''), margin_left=0, margin_right=18)
            if virtual_address[1]:
                x, y, w, h = option.rect.getRect()
                icon_rect = QtCore.QRect(x + w - 18, y + (h-16)/2, 16, 16)
                if virtual_address[0] == 'email':
                    icon = Icon('tango/16x16/apps/internet-mail.png').getQPixmap()
                    painter.drawPixmap(icon_rect, icon)
                elif virtual_address[0] == 'phone':
                    icon = Icon('tango/16x16/devices/phone.png').getQPixmap()
                    painter.drawPixmap(icon_rect, icon)
                elif virtual_address[0] == 'fax':
                    icon = Icon('tango/16x16/devices/printer.png').getQPixmap()
                    painter.drawPixmap(icon_rect, icon)
                elif virtual_address[0] == 'mobile':
                    icon = Icon('tango/16x16/devices/mobile.png').getQPixmap()
                    painter.drawPixmap(icon_rect, icon)
                elif virtual_address[0] == 'im':
                    icon = Icon('tango/16x16/places/instant-messaging.png').getQPixmap()
                    painter.drawPixmap(icon_rect, icon)
                elif virtual_address[0] == 'pager':
                    icon = Icon('tango/16x16/devices/pager.png').getQPixmap()
                    painter.drawPixmap(icon_rect, icon)  
                else:
                #if virtual_adress[0] == 'telephone':
                    icon = Icon('tango/16x16/apps/preferences-desktop-sound.png').getQPixmap()
                    painter.drawPixmap(icon_rect, icon)

        painter.restore()



'''

Created on Sep 25, 2009

@author: Erik De Rijcke

'''
from elixir.entity import Entity, EntityMeta

from sqlalchemy.types import Date, Unicode
from elixir.fields import Field
from elixir.options import using_options
from elixir.relationships import ManyToOne, OneToMany
from camelot.model.authentication import end_of_times
from camelot.view.elixir_admin import EntityAdmin
from camelot.types import Code
import datetime

def type_3_status( statusable_entity, metadata, collection, verbose_entity_name = None ):
    '''
    Creates a new type 3 status related to the given entity
    .. statusable_entity:: A string referring to an entity.
    '''
    t3_status_name = statusable_entity + '_status'
    t3_status_type_name = statusable_entity + '_status_type'

    class Type3StatusTypeMeta( EntityMeta ):
        def __new__( cls, classname, bases, dictionary ):
            return EntityMeta.__new__( cls, t3_status_type_name,
                                 bases, dictionary )
        def __init__( self, classname, bases, dictionary ):
            EntityMeta.__init__( self, t3_status_type_name,
                                  bases, dictionary )

    class Type3StatusType( Entity, ):
        using_options( tablename = t3_status_type_name.lower(), metadata=metadata, collection=collection )
        __metaclass__ = Type3StatusTypeMeta

        code = Field( Code( parts = ['>AAAA'] ), index = True,
                           required = True, unique = True )
        description = Field( Unicode( 40 ), index = True )

        def __unicode__( self ):
            return 'Status type: %s : %s' % ( '.'.join( self.code ), self.description )

        class Admin( EntityAdmin ):
            list_display = ['code', 'description']
            verbose_name = statusable_entity + ' Status Type'
            if verbose_entity_name is not None:
                verbose_name = verbose_entity_name + ' Status Type'

    class Type3StatusMeta( EntityMeta ):
        def __new__( cls, classname, bases, dictionary ):
            return EntityMeta.__new__( cls, t3_status_name,
                                       bases, dictionary )
        def __init__( self, classname, bases, dictionary ):
            EntityMeta.__init__( self, t3_status_name,
                                 bases, dictionary )

    class Type3Status( Entity, ):
        """
        Status Pattern
        .. attribute:: status_datetime For statuses that occur at a specific point in time
        .. attribute:: status_from_date For statuses that require a date range
        .. attribute:: from_date When a status was enacted or set
        """
        using_options( tablename = t3_status_name.lower(), metadata=metadata, collection=collection )
        __metaclass__ = Type3StatusMeta

        status_datetime = Field( Date, required = False )
        status_from_date = Field( Date, required = False )
        status_thru_date = Field( Date, required = False )
        from_date = Field( Date, required = True, default = datetime.date.today )
        thru_date = Field( Date, required = True, default = end_of_times )

        status_for = ManyToOne( statusable_entity, #required = True,
                                 ondelete = 'cascade', onupdate = 'cascade' )
        classified_by = ManyToOne( t3_status_type_name, required = True,
                                   ondelete = 'cascade', onupdate = 'cascade' )

        class Admin( EntityAdmin ):
            verbose_name = statusable_entity + ' Status'
            verbose_name_plural = statusable_entity + ' Statuses'
            list_display = ['status_from_date', 'status_thru_date', 'classified_by']
            verbose_name = statusable_entity + ' Status'
            if verbose_entity_name is not None:
                verbose_name = verbose_entity_name + ' Status'

        def __unicode__( self ):
            return u'Status'

    return t3_status_name

def entity_type( typable_entity, metadata, collection, verbose_entity_name = None ):
    '''
    Creates a new type related to the given entity.
    .. typeable_entity:: A string referring to an entity.
    '''
    type_name = typable_entity + '_type'

    class TypeMeta( EntityMeta ):
        def __new__( cls, classname, bases, dictionary ):
            return EntityMeta.__new__( cls, type_name,
                                       bases, dictionary )
        def __init__( self, classname, bases, dictionary ):
            EntityMeta.__init__( self, type_name,
                                 bases, dictionary )

    class Type( Entity ):
        using_options( tablename = type_name.lower(), metadata=metadata, collection=collection )
        __metaclass__ = TypeMeta

        type_description_for = OneToMany( typable_entity )
        description = Field( Unicode( 48 ), required = True )

        class Admin( EntityAdmin ):
            verbose_name = typable_entity + ' Type'
            list_display = ['description', ]
            verbose_name = typable_entity + ' Type'
            if verbose_entity_name is not None:
                verbose_name = verbose_entity_name + ' Status'

        def __unicode__( self ):
            return u'Type: %s' % ( self.description )

    return type_name




"""Camelot includes a number of Qt delegates, most of them are used as default
delegates for the various sqlalchemy and camelot field types.

Some delegates take specific arguments into account for their construction.
All :attr:`field_attributes` specified for a certain field will be propagated
towards the constructor of the delegate.
"""

from delegatemanager import DelegateManager
from booldelegate import BoolDelegate
from codedelegate import CodeDelegate
from colordelegate import ColorDelegate
from coloredfloatdelegate import ColoredFloatDelegate
from comboboxdelegate import ComboBoxDelegate
from currencydelegate import CurrencyDelegate
from customdelegate import CustomDelegate
from datedelegate import DateDelegate
from datetimedelegate import DateTimeDelegate
from enumerationdelegate import EnumerationDelegate
from filedelegate import FileDelegate
from floatdelegate import FloatDelegate
from imagedelegate import ImageDelegate
from integerdelegate import IntegerDelegate
from intervalsdelegate import IntervalsDelegate
from manytomanydelegate import ManyToManyDelegate
from many2onedelegate import Many2OneDelegate
from one2manydelegate import One2ManyDelegate
from manytoonechoicesdelegate import ManyToOneChoicesDelegate
from plaintextdelegate import PlainTextDelegate
from richtextdelegate import RichTextDelegate
from stardelegate import StarDelegate
from texteditdelegate import TextEditDelegate
from timedelegate import TimeDelegate
from virtualaddressdelegate import VirtualAddressDelegate
from smileydelegate import SmileyDelegate




from PyQt4 import QtGui, QtCore

class AbstractManyToOneEditor(object):
    """Helper functions for implementing a `ManyToOneEditor`, to be used in the
  `ManyToOneEditor` and in the `ManyToManyEditor`
  """
    
    def createSelectView(self):
        #search_text = unicode(self.search_input.text())
        search_text = ''
        admin = self.admin
        query = self.admin.entity.query
        
        class SelectDialog(QtGui.QDialog):
            def __init__(self, parent):
                super(SelectDialog, self).__init__(parent)
                self.entity_selected_signal = QtCore.SIGNAL("entity_selected")
                layout = QtGui.QVBoxLayout()
                layout.setMargin(0)
                layout.setSpacing(0)
                self.setWindowTitle('Select %s'%admin.get_verbose_name())
                self.select = admin.create_select_view(query,
                                                       parent=parent,
                                                       search_text=search_text)
                layout.addWidget(self.select)
                self.setLayout(layout)
                self.connect(self.select, self.select.entity_selected_signal, self.selectEntity)
        
            def selectEntity(self, entity_instance_getter):
                self.emit(self.entity_selected_signal, entity_instance_getter)
                self.close()
        
        selectDialog = SelectDialog(self)
        self.connect(selectDialog, selectDialog.entity_selected_signal, self.selectEntity)
        selectDialog.exec_()
        
    def selectEntity(self, entity_instance_getter):
        #raise Exception('Not implemented')
        raise NotImplementedError



# -*- coding: utf8 -*-

import logging

logger = logging.getLogger( 'camelot.core.files.storage' )

from camelot.view.model_thread import model_function

class StoredFile( object ):
    """Helper class for the File field type.
    Stored file objects can be used within the GUI thread, as none of
    its methods should block.
    """

    def __init__( self, storage, name ):
        """
    :param storage: the storage in which the file is stored
    :param name: the key by which the file is known in the storage"""
        self.storage = storage
        self.name = name

    @property
    def verbose_name( self ):
        """The name of the file, as it is to be displayed in the GUI"""
        return self.name

    def __unicode__( self ):
        return self.verbose_name

class StoredImage( StoredFile ):
    """Helper class for the Image field type Class linking an image and the
    location and filename where the image is stored"""

    @model_function
    def checkout_image( self ):
        """Checkout the image from the storage, and return a QImage"""
        from PyQt4.QtGui import QImage
        p = self.storage.checkout( self )
        return QImage( p )

    @model_function
    def checkout_thumbnail( self, width, height ):
        """Checkout a thumbnail for this image form the storage
        :return: a QImage"""
        from PyQt4.QtCore import Qt
        original_image = self.checkout_image()
        return original_image.scaled( width, height, Qt.KeepAspectRatio )

class Storage( object ):
    """Helper class that opens and saves StoredFile objects
  The default implementation stores files in the settings.CAMELOT_MEDIA_ROOT
  directory.  The storage object should only be used within the model thread,
  as all of it's methods might block.

  The methods of this class don't verify if they are called on the model
  thread, because these classes can be used on the server as well.
    """

    def __init__( self, upload_to = '', stored_file_implementation = StoredFile ):
        """
    :param upload_to: the sub directory in which to put files
    :param stored_file_implementation: the subclass of StoredFile to be used when
    checking out files from the storage
    """
        import settings
        import os
        self.upload_to = os.path.join( settings.CAMELOT_MEDIA_ROOT, upload_to )
        self.stored_file_implementation = stored_file_implementation
        try:
            if not os.path.exists( self.upload_to ):
                os.makedirs( self.upload_to )
        except Exception, e:
            logger.warn( 'Could not access or create path %s, files will be unreachable' % self.upload_to, exc_info = e )

    def exists( self, name ):
        """True if a file exists given some name"""
        import os
        os.path.exists( self.path( name ) )

    def path( self, name ):
        """The local filesystem path where the file can be opened using Python’s standard open"""
        import os
        return os.path.join( self.upload_to, name )

    def checkin( self, local_path ):
        """Check the file pointed to by local_path into the storage, and
        return a StoredFile"""
        import tempfile
        import shutil
        import os
        root, extension = os.path.splitext( os.path.basename( local_path ) )
        ( handle, to_path ) = tempfile.mkstemp( suffix = extension, prefix = root, dir = self.upload_to, text = 'b' )
        os.close( handle )
        logger.debug( 'copy file from %s to %s', local_path, to_path )
        shutil.copy( local_path, to_path )
        return self.stored_file_implementation( self, os.path.basename( to_path ) )

    def checkin_stream( self, prefix, suffix, stream ):
        """Check the datastream in as a file into the storage
        :param prefix: the prefix to use for generating a file name
        :param suffix: the suffix to use for generating a filen name, eg '.png'
        :return: a StoredFile"""
        import tempfile
        import os
        ( handle, to_path ) = tempfile.mkstemp( suffix = suffix, prefix = prefix, dir = self.upload_to, text = 'b' )
        file = os.fdopen( handle, 'wb' )
        file.write( stream.read() )
        file.flush()
        file.close()
        return self.stored_file_implementation( self, os.path.basename( to_path ) )

    def checkout( self, stored_file ):
        """Check the file pointed to by the local_path out of the storage and return
    a local filesystem path where the file can be opened"""
        import os
        return os.path.join( self.upload_to, stored_file.name )

    def checkout_stream( self, stored_file ):
        """Check the file stored_file out of the storage as a datastream
        :return: a file object
        """
        import os
        return open( os.path.join( self.upload_to, stored_file.name ), 'rb' )

    def delete( self, name ):
        pass

class S3Storage( object ):
    """Helper class that opens and saves StoredFile objects into Amazon S3.

    these attibutes need to be set in your settings for S3Storage to work :
     * AWS_ACCESS_KEY_ID = '<INSERT YOUR AWS ACCESS KEY ID HERE>'
     * AWS_SECRET_ACCESS_KEY = '<INSERT YOUR AWS SECRET ACCESS KEY HERE>'
     * AWS_BUCKET_NAME = 'camelot'
     * AWS_LOCATION = S3.Location.DEFAULT
    """

    def __init__( self, upload_to = '', stored_file_implementation = StoredFile ):
        # try to work around bug S3 code which uses bad names of days
        # http://code.google.com/p/boto/issues/detail?id=140
        # but workaround doesn't work :(
        #import locale
#    locale.setlocale(locale.LC_TIME, 'en_US.utf8')
#    print 'create S3 storage'
        import settings
        import S3
        self.upload_to = upload_to
        conn = S3.AWSAuthConnection( settings.AWS_ACCESS_KEY_ID, settings.AWS_SECRET_ACCESS_KEY )
        _generator = S3.QueryStringAuthGenerator( settings.AWS_ACCESS_KEY_ID, settings.AWS_SECRET_ACCESS_KEY )
        if ( conn.check_bucket_exists( settings.AWS_BUCKET_NAME ).status == 200 ):
            pass
        else:
            conn.create_located_bucket( settings.AWS_BUCKET_NAME, settings.AWS_LOCATION ).message



#  ============================================================================
#
#  Copyright (C) 2007-2008 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file LICENSE.GPL included in the packaging of
#  this file.  Please review the following information to ensure GNU
#  General Public Licensing requirements will be met:
#  http://www.trolltech.com/products/qt/opensource.html
#
#  If you are unsure which license is appropriate for your use, please
#  review the following information:
#  http://www.trolltech.com/products/qt/licensing.html or contact
#  project-camelot@conceptive.be.
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Utility functions"""

from PyQt4 import QtCore
from PyQt4.QtCore import QCoreApplication


def create_constant_function(constant):
    return lambda:constant


def variant_to_pyobject(qvariant=None):
    """Try to convert a QVariant to a python object as good
    as possible"""
    import datetime
    if not qvariant:
        return None
    if qvariant.isNull():
        return None
    type = qvariant.type()
    if type == QtCore.QVariant.String:
        value = unicode(qvariant.toString())
    elif type == QtCore.QVariant.Date:
        value = qvariant.toDate()
        value = datetime.date(year=value.year(),
                              month=value.month(),
                              day=value.day())
    elif type == QtCore.QVariant.Int:
        value = int(qvariant.toInt()[0])
    elif type == QtCore.QVariant.LongLong:
        value = int(qvariant.toLongLong()[0])      
    elif type == QtCore.QVariant.Double:
        value = float(qvariant.toDouble()[0])
    elif type == QtCore.QVariant.Bool:
        value = bool(qvariant.toBool())
    elif type == QtCore.QVariant.Time:
        value = qvariant.toTime()
        value = datetime.time(hour = value.hour(),
                              minute = value.minute(),
                              second = value.second())
    elif type == QtCore.QVariant.DateTime:
        value = qvariant.toDateTime()
        value = value.toPyDateTime ()
    else:
        value = qvariant.toPyObject()
      
    return value

#
# Global dictionary containing all user defined translations in the
# current locale
#  
_translations_ = {}

def load_translations():
    """Fill the global dictionary of translations with all data from the
    database, to be able to do fast gui thread lookups of translations"""
    language = unicode(QtCore.QLocale().name())
    from camelot.model.i18n import Translation
    tls = Translation.query.filter(Translation.language==language)
    tls = tls.filter(Translation.value!=None).all()
    for t in tls:
        if t.value:
            _translations_[t.source] = t.value
  
def ugettext(string_to_translate):
    """Translate the string_to_translate to the language of the current locale.
    This is a two step process.  First the function will try to get the
    translation out of the Translation entity, if this is not successfull, the
    function will ask QCoreApplication to translate string_to_translate 
    (which tries to get the translation from the .po files)"""
    result = _translations_.get(string_to_translate, None)
    if not result:
        result = unicode(QCoreApplication.translate('', string_to_translate))
    return result
  
class ugettext_lazy(object):
  
    def __init__(self, string_to_translate):
        self._string_to_translate = string_to_translate
      
    def __str__(self):
        return ugettext(self._string_to_translate)
    
    def __unicode__(self):
        return ugettext(self._string_to_translate)




from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

from customeditor import CustomEditor, editingFinished
from camelot.view.art import Icon
from camelot.core.utils import ugettext as _

class FileEditor(CustomEditor):
    """Widget for editing File fields"""
  
    filter = """All files (*)"""
    
    def __init__(self, parent=None, storage=None, editable=True, **kwargs):
        CustomEditor.__init__(self, parent)
        self.storage = storage
        self.editable = editable
        self.document_pixmap = Icon('tango/16x16/mimetypes/x-office-document.png').getQPixmap()
        self.clear_icon = Icon('tango/16x16/actions/edit-delete.png').getQIcon()
        self.new_icon = Icon('tango/16x16/actions/list-add.png').getQIcon()
        self.open_icon = Icon('tango/16x16/actions/document-open.png').getQIcon()
        self.saveas_icon = Icon('tango/16x16/actions/document-save-as.png').getQIcon()
        self.value = None
        self.setup_widget()
        if self.editable:
            self.setAcceptDrops(True)
    
    def setup_widget(self):
        """Called inside init, overwrite this method for custom
        file edit widgets"""
        self.layout = QtGui.QHBoxLayout()
        self.layout.setSpacing(0)
        self.layout.setMargin(0)
        
        # Clear button
        self.clear_button = QtGui.QToolButton()
        self.clear_button.setFocusPolicy(Qt.ClickFocus)
        self.clear_button.setIcon(self.clear_icon)
        self.clear_button.setToolTip(_('delete file'))
        self.clear_button.setAutoRaise(True)
        self.clear_button.setEnabled(self.editable)
        self.connect(self.clear_button,
                     QtCore.SIGNAL('clicked()'),
                     self.clear_button_clicked)
        
        # Open button
        self.open_button = QtGui.QToolButton()
        self.open_button.setFocusPolicy(Qt.ClickFocus)
        self.open_button.setIcon(self.new_icon)
        self.open_button.setToolTip(_('add file'))
        self.open_button.setEnabled(self.editable)
        self.connect(self.open_button,
                     QtCore.SIGNAL('clicked()'),
                     self.open_button_clicked)
        self.open_button.setAutoRaise(True)
        
        # Filename
        self.filename = QtGui.QLineEdit(self)
        self.filename.setEnabled(self.editable  )
        self.filename.setReadOnly(not self.editable)
        
        # Setup layout
        self.document_label = QtGui.QLabel(self)
        self.document_label.setPixmap(self.document_pixmap)
        self.document_label.setEnabled(self.editable)
        self.layout.addWidget(self.document_label)
        self.layout.addWidget(self.filename)
        self.layout.addWidget(self.clear_button)
        self.layout.addWidget(self.open_button)
        self.setLayout(self.layout)      
    
    def set_value(self, value):
        value = CustomEditor.set_value(self, value)
        self.value = value
        if value:
            self.filename.setText(value.verbose_name)
            self.open_button.setIcon(self.open_icon)
            self.open_button.setToolTip(_('open file'))
        else:
            self.filename.setText('')
            self.open_button.setIcon(self.new_icon)
            self.open_button.setToolTip(_('add file'))
        return value
        
    def get_value(self):
        return CustomEditor.get_value(self) or self.value
    
    def set_enabled(self, editable=True):
        self.clear_button.setEnabled(editable)
        self.open_button.setEnabled(editable)
        self.filename.setEnabled(editable)
        self.document_label.setEnabled(editable)
    
    def stored_file_ready(self, stored_file):
        """Slot to be called when a new stored_file has been created by
        the storeage"""
        self.set_value(stored_file)
        self.emit(editingFinished)        
        
    def open_button_clicked(self):
        from camelot.view.storage import open_stored_file, create_stored_file
        if not self.value:
            create_stored_file(self, self.storage, self.stored_file_ready, filter=self.filter)
        else:
            open_stored_file(self, self.value)
    
    def clear_button_clicked(self):
        self.value = None
        self.emit(editingFinished)

    #
    # Drag & Drop
    #
    def dragEnterEvent(self, event):
        event.acceptProposedAction()
    
    def dragMoveEvent(self, event):
        event.acceptProposedAction()
    
    def dropEvent(self, event):
        if event.mimeData().hasUrls():
            url = event.mimeData().urls()[0]
            filename = url.toLocalFile()
            if filename != '':
                from camelot.view.storage import SaveFileProgressDialog
                from camelot.view.model_thread import post
                progress = SaveFileProgressDialog()
                
                def checkin():
                    stored_file = self.storage.checkin(str(filename))
                    return lambda:self.stored_file_ready(stored_file)

                post(checkin, progress.finish)
                progress.exec_()
